/*
**	  GPU Raytrace - Diploma thesis, Czech technical university
**	  Copyright (C) 2008 Martin Zlatuska
**
**	  This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdlib.h>
#include <stdio.h>

#include <limits>

#include <string.h>

#include <xmmintrin.h>

#include "Object.h"
#include "Material.h"

#include "PlyLoader.h"

#include "Transformations/SceneNode.h"

using sse::operator==;

static const unsigned addDummy = SceneLoader::instance()->addLoader(new PlyLoader,"ply");	//Register the file loader with SceneLoader wrapper.

PlyLoader::PlyLoader(void) {
};

PlyLoader::~PlyLoader(void) {
}

Object* PlyLoader::loadFile(const char *fileName) {
	///////////////////////////
	//Read whole file in memory
	FILE*  input;
	size_t fileSize;

	input = fopen(fileName,"r");

	fseek(input,0,SEEK_END);
	fileSize = ftell(input);
	fseek(input,0,SEEK_SET);

	char* buffer = new char[fileSize+1];

	fileSize = fread(buffer,1,fileSize,input);
	fclose(input);

	buffer[fileSize]=0;

	char* cur = strtok(buffer,"\n");

	///////////////////////////
	//Simplyfied header parsing.
	//Only recognize some of the header informations and ignore everything else

	int vertexCount;
	int faceCount;

	
	while(true) {
		cur = strtok(0,"\n");

		if(!strncmp(cur,"element vertex",14)) {
			vertexCount = strtol(cur + 14,&cur,10);
		} else if (!strncmp(cur,"element face",12)) {
			faceCount   = strtol(cur + 12,&cur,10);
		} else if (!strcmp(cur,"end_header") )
			break;
	};

	while(*cur != 0) cur++;
	cur++;

	//Do bounding box computation during parsing to reduce bandwidth.
	//Init initial values to oposite extreme.
	__m128 minimum,maximum;
	//minimum = _mm_load_ps(minimumInit3);
	//maximum = _mm_load_ps(maximumInit3);
	minimum = minimumInit3;
	maximum = maximumInit3;
	
	//Create new object and allocate space for vertices and faces
	Object *object   = new Object;
	object->vertices = new Vertex[vertexCount];
	object->normal	 = new Vertex[vertexCount];
	object->faces	 = new Triangle[faceCount];
	object->vertexCount = vertexCount;
	object->facesCount	= faceCount;
	object->materials.push_back(0);

	float ALIGN16 temp[4];
	temp[3]=0.0f;
	for(int i=0;i<vertexCount;i++) {
		//Workaround the Microsoft runtime library performance issue. The itoa calls strlen
		//every time it's used. So using it on long string results in extreme slowdown.
		//Fix by choping the string to smaller ones. Since we will not use it anymore there
		//is no need to restore linebreak after sscanf
		char* next = strchr(cur,'\n');
		*next = 0;

		//Parse vertex coordinates
		object->vertices[i].w = 1.0f;
		sscanf(cur,"%f %f %f",&object->vertices[i].x,&object->vertices[i].y,&object->vertices[i].z);

		//Update bounding box coordinates
		__m128 current = _mm_load_ps(&object->vertices[i].x);
		minimum = _mm_min_ps(minimum,current);
		maximum = _mm_max_ps(maximum,current);

		object->normal[i].vec = zeroesInit;
		
		cur = next+1;
	};

	for(int i=0;i<faceCount;i++) {
		//The same performace fix.
		char* next = strchr(cur,'\n');
		*next = 0;

		//Parse triangle indexes
		unsigned int count;
		sscanf(cur,"%u %u %u %u",&count,&object->faces[i].a,&object->faces[i].b,&object->faces[i].c);
		object->faces[i].material = 0;

		cur = next+1;
		
	};

	//The used models have no normal informations. Calculate every vertex normal from normals of it's tirangles.
	//Treat object as smooth without sharp edeges.
	for(int i=0;i<faceCount;i++) {
		unsigned v1,v2,v3;
		v1 = object->faces[i].a;
		v2 = object->faces[i].b;
		v3 = object->faces[i].c;

		SPVector a = object->vertices[v1];
		SPVector b = object->vertices[v2];
		SPVector c = object->vertices[v3];

		b-=a;
		c-=a;

		SPVector normal = b % c;

		//Some of the triangles are degenerated and are actualy lines. In that case we must no try to normalize
		//the normal(infinite result) and add it.
		__m128 mask = normal.vec == zeroesInit;
		char  xmask = _mm_movemask_ps(mask);

		if(xmask!=0xF) {
			//The triangle size infulences the normal lenght. To prevent this produving wrong output, normalize before add.
			normal.Normalize();
			object->normal[v1]+= normal;
			object->normal[v2]+= normal;
			object->normal[v3]+= normal;
		};
	};

	//The added normals must be renormalize.
	for(int i=0;i<vertexCount;i++)
		object->normal[i].Normalize();

	object->minimum.vec = minimum;
	object->maximum.vec = maximum;

	delete buffer;

	return object;
}



