/*
**	  GPU Raytrace - Diploma thesis, Czech technical university
**	  Copyright (C) 2008 Martin Zlatuska
**
**	  This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include "Camera.h"
#include "SceneLights.h"
#include "Partitoning/UniformGrid.h"
#include "Partitoning/KdTree.h"
#include "Partitoning/Bvh.h"

#include "RenderMode.h"

enum RenderTarget {
	MEMORY,
	PBO
};

//! Parent class for all other CUDA wrappers used to launch the raytracing.
/**
*	Wrapper class around cuda function call. It's purpose is to make the high level passing of 
*	acceleration structure to cuda function easier. The wrapper is supposed to be imutable in
*	respect to the acceleration structure. Therefore the acceleration structure is assigned 
*	during object construction, wrapper takes ownership of the structure and it is deleted during
*	wrapper destruction.
*/
class CudaRenderer {
protected:
	float3 minimum;		//!< Minimum extent of the structure bounding box.
	float3 maximum;		//!< Maximum extent of the structure bounding box.

	RenderTarget	renderTarget;	//!< The current rendergin target.

	unsigned width;		//!< Width of rendered image.
	unsigned height;	//!< Height of rendered image.

	int			  pbo;		//!< OpenGL buffer used to store renderer image when view on screen.
	unsigned int* buffer;	//!< Frame buffer for raytraced image in video card memory.
	unsigned int* copy;		//!< Host memory used as a copy of video card frame buffer.

public:
	CudaRenderer(unsigned width,unsigned height,int pbo);
	virtual ~CudaRenderer();

	//! Set new size of the rendered image.
	/**
	*	@param newWidth	 New width of the image.
	*	@param newHeight New height of the image.
	*/
	void resize(unsigned newWidth,unsigned newHeight);

	//! Sets new render target.
	/**
	*	Change the PBO used as buffer or switch to rendering to memory.
	*	@param newPbo	New OpenGL PBO to used as buffer, -1 for render to memory
	*/
	void setRenderTarget(unsigned newPbo=-1);

	//! Return the pointer to host memory copy of the buffer.
	unsigned int* getBuffer() {return copy;};

	//! Raycast the image.
	/**
	*	@param camera	The camera used to render the image.
	*	@param lights   The light in the scene.
	*	@param mode		The render mode(primary rays only,primary+shadow,primary+shadow+reflect).
	*/
	virtual float render(Camera& camera,SceneLights &lights,RenderMode mode)=0;

	float getRenderDistance() {
		return std::max(maximum.x - minimum.x,maximum.y - minimum.y);
	};
};

//! Wrapper for raytracing using uniform grid acceleration structure.
class CudaRenderGrid:public CudaRenderer {
	
	float  cellSize;	//!< Size of the one cell.
	uint3  dim;			//!< Number of cells in each dimension.

	UniformGrid* grid;			//!< Pointer to the acceleration structure.

	//Data obtained from the acceleration structure
	cudaArray*	  cudaGrid;
	uint4*		  triangles;
	cudaArray*	  vertices;
	cudaArray*	  normals;
	Material*	  materials;

public:
	//! Prepare uniform grid for rendering.
	/**
	*	@param width	Initial width of the image.
	*	@param height	Initial height of the image.
	*	@param grid		The uniform grid used for rendering. The wrapper assumes ownership of the structure and deletes it during destruction.
	*/
	CudaRenderGrid(unsigned width,unsigned height,UniformGrid *grid,int pbo=-1);
	~CudaRenderGrid();

	float render(Camera& camera,SceneLights &lights,RenderMode mode);
};

class CudaRenderKd:public CudaRenderer {
	KdTree* tree;	//!< Pointer to the acceleration structure.

	//Data obtained from the acceleration structure
	cudaArray*	  cudaTree;
	uint4*		  triangles;
	cudaArray*	  vertices;
	cudaArray*	  normals;
	Material*	  materials;

public:
	//! Prepare kd-tree for rendering.
	/**
	*	@param width	Initial width of the image.
	*	@param height	Initial height of the image.
	*	@param tree		The kd-tree used for rendering. The wrapper assumes ownership of the structure and deletes it during destruction.
	*/
	CudaRenderKd(unsigned width,unsigned height,KdTree *tree,int pbo=-1);
	~CudaRenderKd();

	float render(Camera& camera,SceneLights &lights,RenderMode mode);
};

class CudaRenderBvh:public CudaRenderer {
	Bvh *tree;		//!< Pointer to the acceleration structure.

	//Data obtained from the acceleration structure
	unsigned*	  bvhTree;
	uint4*		  triangles;
	cudaArray*	  vertices;
	cudaArray*	  normals;
	Material*	  materials;

public:
	//! Prepare kd-tree for rendering.
	/**
	*	@param width	Initial width of the image.
	*	@param height	Initial height of the image.
	*	@param tree		The BVH tree used for rendering. The wrapper assumes ownership of the structure and deletes it during destruction.
	*/
	CudaRenderBvh(unsigned width,unsigned height,Bvh *tree,int pbo=-1);
	~CudaRenderBvh();

	float render(Camera& camera,SceneLights &lights,RenderMode mode);
};