/*
**	  GPU Raytrace - Diploma thesis, Czech technical university
**	  Copyright (C) 2008 Martin Zlatuska
**
**	  This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "CameraCuda.h"
#include "Lights.h"
#include "MaterialCuda.h"


texture<float4, 2, cudaReadModeElementType> vertexTexture;
texture<float4, 2, cudaReadModeElementType> normalTexture;

__constant__ unsigned lightsCount;
__constant__ Light    lights[8];

//Get face normal at given intersection point
__device__ void catchNormal(uint4* triangles, int hitTriangle,float u,float v,float3& normal,unsigned &material) {
	//Normals are stored in 2D texture(Due to the 1D size limitation) with width of 256. Convert linear indexes to 2D coordinates.
	unsigned x1 = triangles[hitTriangle].y &  255;
	unsigned y1 = triangles[hitTriangle].y >> 8;
	unsigned x2 = triangles[hitTriangle].z &  255;
	unsigned y2 = triangles[hitTriangle].z >> 8;
	unsigned x3 = triangles[hitTriangle].w &  255;
	unsigned y3 = triangles[hitTriangle].w >> 8;
	material	= triangles[hitTriangle].x;

	//Read normals from texture
	float3 normal0 = float4to3(tex2D(normalTexture,x1,y1));
	float3 normal1 = float4to3(tex2D(normalTexture,x2,y2));
	float3 normal2 = float4to3(tex2D(normalTexture,x3,y3));

	//The intesection point is given is barycentric coordinates and is therefore weighted average of vertices.
	//The third missing coordinate can be computed from the supplied two as sum must be 1.
	float w = 1.0f - u - v;
	normal0*=w;
	normal1*=u;
	normal2*=v;

	//sum normals and renormalize
	normal0+=normal1;
	normal0+=normal2;
	normalize(normal0);

	normal = normal0;
};

__device__ void shade(float3 &color,bool shadow,unsigned light,
					  Material* materials,unsigned material,
					  float currentVal,float refletSum,
					  float3 &normal,float3 &dir,float3 &reflectedDir) {
	//The standard Phong shading model
	normalize(dir);

	//The ambient light.
	color += materials[material].ambient * lights[light].kAmbient * currentVal;

	//Do diffuse and specular only if not in shadow
	if(!shadow) {
		float coef;

		//Diffuse light
		coef = abs(dot(normal,dir));
		if(coef >= 0.0f)		
			//This is irelevant as coef is now always positive. But in current compiler increase register usage if removed(dont know why);
			color += materials[material].diffuse * lights[light].kDiffuse * coef * currentVal;

		//Specular light
		normalize(reflectedDir);
		coef = -dot(dir,reflectedDir);
		if(coef > 0.0f) {
			coef = powf(coef,materials[material].shinines);
			color += materials[material].specular * lights[light].kDiffuse * coef * refletSum;
		};
	};
};

__device__ void shadePrimary(float3 &color,unsigned light,
							 Material* materials,unsigned material,
						     float3 &normal,float3 &dir,float3 &reflectedDir) {
	//The standard Phong shading model
	normalize(dir);

	//The ambient light.
	color += materials[material].ambient * lights[light].kAmbient;
	
	float coef;
	//Diffuse light
	coef = abs(dot(normal,dir));
	if(coef >= 0.0f)		
		//This is irelevant as coef is now always positive. But in current compiler increase register usage if removed(dont know why);
		color += materials[material].diffuse * lights[light].kDiffuse * coef;

	//Specular light
	normalize(reflectedDir);
	coef = -dot(dir,reflectedDir);
	if(coef > 0.0f) {
		coef   = powf(coef,materials[material].shinines);
		color += materials[material].specular * lights[light].kDiffuse * coef;
	};
};

__device__ void storeInteger(float3 &color,unsigned int* buffer,unsigned startX,unsigned startY,unsigned width) {
/*	unsigned short x = blockIdx.x*8 + threadIdx.x;
	unsigned short y = blockIdx.y*8 + threadIdx.y;
	unsigned index = __mul24(y*8,gridDim.x) + x;

	color.x = min(color.x,1.0f);
	color.y = min(color.y,1.0f);
	color.z = min(color.z,1.0f);

	color *= 255.0f;
	uint3 iColor;
	iColor.x = color.x;
	iColor.y = color.y;
	iColor.z = color.z;

	unsigned val = iColor.z<<16 | iColor.y<<8 | iColor.x;
	buffer[index] = val;*/

	unsigned short x = startX + blockIdx.x*8 + threadIdx.x;
	unsigned short y = startY + blockIdx.y*8 + threadIdx.y;
	unsigned index = __mul24(y,width) + x;

	color.x = min(color.x,1.0f);
	color.y = min(color.y,1.0f);
	color.z = min(color.z,1.0f);

	color *= 255.0f;
	uint3 iColor;
	iColor.x = color.x;
	iColor.y = color.y;
	iColor.z = color.z;

	unsigned val = iColor.z<<16 | iColor.y<<8 | iColor.x;
	buffer[index] = val;
};

__device__
void intersectBox(float3 &minimum,float3 &maximum,float3 &eye,float3 &dir,float &tMin,float &tMax) {
	// Kay and Kayjia ray-box intersection test as descibed in "An introduction to Ray Tracing" by A.S.Glassner
	float3 reciprocal = rcp(dir);
	float3 tHitLeft	  = (minimum - eye) * reciprocal;
	float3 tHitRight  = (maximum - eye) * reciprocal;
	
	float3 tHitMin = min(tHitLeft,tHitRight);
	float3 tHitMax = max(tHitLeft,tHitRight);

	tMax = min(min(tHitMax.x,tHitMax.y),tHitMax.z);
	tMin = max(max(tHitMin.x,tHitMin.y),tHitMin.z);
};

__device__
void intersectTexture(uint4* triangles,float3 &eye,float3 &dir,unsigned offset,float &u,float &v,float &t) {
	unsigned x1 = triangles[offset].y &  255;
	unsigned y1 = triangles[offset].y >> 8;
	unsigned x2 = triangles[offset].z &  255;
	unsigned y2 = triangles[offset].z >> 8;
	unsigned x3 = triangles[offset].w &  255;
	unsigned y3 = triangles[offset].w >> 8;

	float3 v0 = float4to3(tex2D(vertexTexture,x1,y1));
	float3 v1 = float4to3(tex2D(vertexTexture,x2,y2));
	float3 v2 = float4to3(tex2D(vertexTexture,x3,y3));

	v2 -= v0;
	v1 -= v0;

	float3 tvec   = eye - v0;
	float3 pvec   = cross(dir,v2);
	float  invDet = 1.0f / dot(v1,pvec);

	u = dot(tvec,pvec)*invDet;

	float3 qvec	  = cross(tvec,v1);
	t = dot(v2,qvec)*invDet;
	v = dot(dir,qvec)*invDet;
};

__device__ void catchNormalIntersection(uint4* triangles, int hitTriangle,float3 &eye,float3 &dir,float3& normal,float& t,unsigned &material) {
	//Normals are stored in 2D texture(Due to the 1D size limitation) with width of 256. Convert linear indexes to 2D coordinates.
	unsigned x1 = triangles[hitTriangle].y &  255;
	unsigned y1 = triangles[hitTriangle].y >> 8;
	unsigned x2 = triangles[hitTriangle].z &  255;
	unsigned y2 = triangles[hitTriangle].z >> 8;
	unsigned x3 = triangles[hitTriangle].w &  255;
	unsigned y3 = triangles[hitTriangle].w >> 8;
	material	= triangles[hitTriangle].x;

	float3 v0 = float4to3(tex2D(vertexTexture,x1,y1));
	float3 v1 = float4to3(tex2D(vertexTexture,x2,y2));
	float3 v2 = float4to3(tex2D(vertexTexture,x3,y3));

	v2 -= v0;
	v1 -= v0;

	float3 tvec   = eye - v0;
	float3 pvec   = cross(dir,v2);
	float  invDet = 1.0f / dot(v1,pvec);
	float  u	  = dot(tvec,pvec)*invDet;
	float3 qvec	  = cross(tvec,v1);
		   t	  = dot(v2  ,qvec)*invDet;
	float  v	  = dot(dir ,qvec)*invDet;

	//Read normals from texture
		   normal  = float4to3(tex2D(normalTexture,x1,y1));
	float3 normal1 = float4to3(tex2D(normalTexture,x2,y2));
	float3 normal2 = float4to3(tex2D(normalTexture,x3,y3));

	//The intesection point is given is barycentric coordinates and is therefore weighted average of vertices.
	//The third missing coordinate can be computed from the supplied two as sum must be 1.
	float w = 1.0f - u - v;
	normal *=w;
	normal1*=u;
	normal2*=v;

	//sum normals and renormalize
	normal+=normal1;
	normal+=normal2;
	normalize(normal);
};