/*
**	  GPU Raytrace - Diploma thesis, Czech technical university
**	  Copyright (C) 2008 Martin Zlatuska
**
**	  This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <float.h>

texture<uint2 , 3, cudaReadModeElementType> gridTexture;

__constant__ float3 steps[8]={{-1.0f,-1.0f,-1.0f},{1.0f,-1.0f,-1.0f},{-1.0f,1.0f,-1.0f},{1.0f,1.0f,-1.0f},
							  {-1.0f,-1.0f, 1.0f},{1.0f,-1.0f, 1.0f},{-1.0f,1.0f, 1.0f},{1.0f,1.0f, 1.0f}};
__constant__ float3 outs[8];
__constant__ float3 limits;

__device__
void intersectTexture(uint4* triangles,float3 &eye,float3 &dir,unsigned offset,float &u,float &v,float &t) {
	unsigned x1 = triangles[offset].y &  255;
	unsigned y1 = triangles[offset].y >> 8;
	unsigned x2 = triangles[offset].z &  255;
	unsigned y2 = triangles[offset].z >> 8;
	unsigned x3 = triangles[offset].w &  255;
	unsigned y3 = triangles[offset].w >> 8;

	float3 v0 = float4to3(tex2D(vertexTexture,x1,y1));
	float3 v1 = float4to3(tex2D(vertexTexture,x2,y2));
	float3 v2 = float4to3(tex2D(vertexTexture,x3,y3));

	v2 -= v0;
	v1 -= v0;

	float3 tvec   = eye - v0;
	float3 pvec   = cross(dir,v2);
	float  invDet = 1.0f / dot(v1,pvec);

	u = dot(tvec,pvec)*invDet;

	float3 qvec	  = cross(tvec,v1);
	t = dot(v2,qvec)*invDet;
	v = dot(dir,qvec)*invDet;
};

__device__ void catchNormalIntersection(uint4* triangles, int hitTriangle,float3 &eye,float3 &dir,float3& normal,float& t,unsigned &material) {
	//Normals are stored in 2D texture(Due to the 1D size limitation) with width of 256. Convert linear indexes to 2D coordinates.
	unsigned x1 = triangles[hitTriangle].y &  255;
	unsigned y1 = triangles[hitTriangle].y >> 8;
	unsigned x2 = triangles[hitTriangle].z &  255;
	unsigned y2 = triangles[hitTriangle].z >> 8;
	unsigned x3 = triangles[hitTriangle].w &  255;
	unsigned y3 = triangles[hitTriangle].w >> 8;
	material	= triangles[hitTriangle].x;

	float3 v0 = float4to3(tex2D(vertexTexture,x1,y1));
	float3 v1 = float4to3(tex2D(vertexTexture,x2,y2));
	float3 v2 = float4to3(tex2D(vertexTexture,x3,y3));

	v2 -= v0;
	v1 -= v0;

	float3 tvec   = eye - v0;
	float3 pvec   = cross(dir,v2);
	float  invDet = 1.0f / dot(v1,pvec);
	float  u	  = dot(tvec,pvec)*invDet;
	float3 qvec	  = cross(tvec,v1);
		   t	  = dot(v2  ,qvec)*invDet;
	float  v	  = dot(dir ,qvec)*invDet;

	//Read normals from texture
		   normal  = float4to3(tex2D(normalTexture,x1,y1));
	float3 normal1 = float4to3(tex2D(normalTexture,x2,y2));
	float3 normal2 = float4to3(tex2D(normalTexture,x3,y3));

	//The intesection point is given is barycentric coordinates and is therefore weighted average of vertices.
	//The third missing coordinate can be computed from the supplied two as sum must be 1.
	float w = 1.0f - u - v;
	normal *=w;
	normal1*=u;
	normal2*=v;

	//sum normals and renormalize
	normal+=normal1;
	normal+=normal2;
	normalize(normal);
};

template<bool output>
__device__ int traverseTexture(float3 &eye,float3 &dir,float3 &minimum ,float3 &maximum,float  cellSize, uint4* triangles) {
	float  cellX,cellY,cellZ;

	//The traversal algorithm needs 6 informative value for its operation. The cell increment value and the coordinates
	//of the cell, at whitch to reminate. This depend on the ray direction and is independent in each direction. This
	//also takes space of 6 register. The possible values are same for all the ray. So instead of stroring them per-ray
	//they are stored in device constant memory as an array and an index to the right value is computed from ray direction.
	//This saves 5 register and allow the primary ray only version to use only 32 registers and achieve higher utilisation and speed.
	int    stepAll = 0;

	float3 tDelta;
	float3 tNext;

	bool hit = true;

	// Kay and Kayjia ray-box intersection test as descibed in "An introduction to Ray Tracing" by A.S.Glassner
	{
		float3 reciprocal = rcp(dir);
		float3 tHitLeft	  = (minimum - eye) * reciprocal;
		float3 tHitRight  = (maximum - eye) * reciprocal;
	
		float3 tHitMin = min(tHitLeft,tHitRight);
		float3 tHitMax = max(tHitLeft,tHitRight);

		float  tMax = min(min(tHitMax.x,tHitMax.y),tHitMax.z);
		float  tMin = max(max(tHitMin.x,tHitMin.y),tHitMin.z);
		hit  = tMin<=tMax && hit;
		
		tMin			 = max(tMin,0.0f);
		float3 intersect = eye + dir*tMin;

		//Convert intersection point to grid cell coordinates
		intersect -= minimum;

		//Due to numerical error, intersection point can be outside grid, prevent negative coordinates.
		intersect.x = max(intersect.x,0.0f);
		intersect.y = max(intersect.y,0.0f);
		intersect.z = max(intersect.z,0.0f);
		intersect *= (1.0f/cellSize);

		//The coordinates can also be outside the grid in other direction, clamp to grid size while rounding 
		//to exact cell index.
		cellX = min(floor(intersect.x),limits.x);
		cellY = min(floor(intersect.y),limits.y);
		cellZ = min(floor(intersect.z),limits.z);

		//Calculate the number of ray parameter change when steping one cell in each dimension
		tDelta.x = cellSize * abs(reciprocal.x);
		tDelta.y = cellSize * abs(reciprocal.y);
		tDelta.z = cellSize * abs(reciprocal.z);
		
		//Calculate the initial next-cell parameter value. Find the distance from the intersection point to
		//the next cell boundary in the respect to the ray direction. Convert this distance to paramtere value and
		//adjust the intersection point value by it.
		float dx = cellX - intersect.x;
		float dy = cellY - intersect.y;
		float dz = cellZ - intersect.z;
		if(dir.x >= 0) {dx+=1; stepAll |= 0x01;};
		if(dir.y >= 0) {dy+=1; stepAll |= 0x02;};
		if(dir.z >= 0) {dz+=1; stepAll |= 0x04;};

		tNext.x = tMin + dx*cellSize*reciprocal.x;
		tNext.y = tMin + dy*cellSize*reciprocal.y;
		tNext.z = tMin + dz*cellSize*reciprocal.z;
	};

	int  hitTriangle = -1;
	float closestHit = FLT_MAX; //Set closest intersection point to "infinity"

	//Traverse grid only if the ray actually hit it
	if(hit) {
		while(hitTriangle==-1) {	
			uint2 cell = tex3D(gridTexture,cellX,cellY,cellZ);	//Catch cell info from texture

			float nextCell = min(min(tNext.x,tNext.y),tNext.z);

			if(cell.y) {	//If current cell is non-empty
				unsigned offset = cell.x;	//Initialize offset to first triangle
				closestHit = FLT_MAX;

				//Loop over all the cell's triangles
				for(int i=0;i<cell.y;i++) {
					float u,v,t;
					intersectTexture(triangles,eye,dir,offset,u,v,t);

					//If the ray hits the triangle and is closer then previus intersection and not to close, mark as hit
					//The comparisons try to awoid some numerical precision problems. The values here are empirical and
					//it seems they work quite right, but it might be necessary to change them later.
					if(v>=-0.00048828125f && u>=-0.00048828125f && u+v<=1.0000304f && t>=0.010070801f && t<=closestHit && t<nextCell+0.00048828125f) {
						closestHit  = t;
						hitTriangle = offset; 
					};

					offset++;
				};
			};

			//The cell side through witch the ray exits is the one with smallest parameter.
			if(tNext.x < tNext.y) {
				if(tNext.x < tNext.z) {
					cellX += steps[stepAll].x;		//Move to next cell, the actual value depends on ray direction (>0 or <0) in respective dimension.
					if(cellX == outs[stepAll].x)	//If we step outside the grid, terminate.
						break;
					tNext.x += tDelta.x;			//Update the next size intersection parameter.
				} else {
					cellZ += steps[stepAll].z;
					if(cellZ == outs[stepAll].z)
						break;
					tNext.z += tDelta.z;
				};
			} else {
				if(tNext.y < tNext.z) {
					cellY += steps[stepAll].y;	
					if(cellY == outs[stepAll].y)
						break;
					tNext.y += tDelta.y;
				} else {
					cellZ += steps[stepAll].z;
					if(cellZ == outs[stepAll].z) 
						break;
					tNext.z += tDelta.z;
				};
			};
		};
	};

	if(output)
		return hitTriangle;
	else
		return hitTriangle == -1 ? __float_as_int(-1.0f) :  __float_as_int(closestHit);
};

__global__ void renderGridPSR(float3 eye		,float3 upperLeft, float3 dx, float3 dy,unsigned short startX,unsigned short startY,unsigned width,
							  float3 minimum ,float3 maximum, float  cellSize, 
							  uint4* triangles,Material* materials,
							  unsigned int* buffer) {
	unsigned	reflectLimit = 2;		//Limit number of reflection to fixed number. Might change to paramtere later
	float		refletSum	 = 1.0f;	//

	//Compute pixel coordinates. The grid is composed of 8x8 blocks.
	unsigned short x  = startX + blockIdx.x*8 + threadIdx.x;
	unsigned short y  = startY + blockIdx.y*8 + threadIdx.y;

	//if(!(blockIdx.x == 39 && blockIdx.y ==42)) return;
	
	//Compute ray direction for current pixel. To achive right traversal, use higher precision intrinsic.
	float3 dir;
	dir.x = upperLeft.x + __fmul_rn(dx.x,x) + __fmul_rn(dy.x,y);
	dir.y = upperLeft.y + __fmul_rn(dx.y,x) + __fmul_rn(dy.y,y);
	dir.z = upperLeft.z + __fmul_rn(dx.z,x) + __fmul_rn(dy.z,y);

	float3 color = make_float3(0.0f,0.0f,0.0f);	//Start with black pixel color

	float3 origin = eye;
	while(true) {
		int hitTriangle = traverseTexture<true>(origin,dir,minimum,maximum,cellSize,triangles);

		unsigned material;
		float3   intersectionPoint;
		float3	 reflectedDir;
		float3   normal;

		if(hitTriangle >=0) {
			float  t;
			
			catchNormalIntersection(triangles,hitTriangle,origin,dir,normal,t,material);

			intersectionPoint = origin + dir*t;							//Compute intersection point of ray and triangle.
			reflectedDir	  = dir    - normal*dot(normal,dir)*2.0f;	//Reflected ray computation
		};

		unsigned shadowPass = 0;

		if(hitTriangle >= 0) {
			while(shadowPass < lightsCount) {
				origin = lights[shadowPass].possition;
				dir    = intersectionPoint - origin;

				float t = __int_as_float(traverseTexture<false>(origin,dir,minimum,maximum,cellSize,triangles));
		
				bool  shadow = t>=0.0f && t<0.999f;
				float currentVal = refletSum*(1.0f - materials[material].reflectivity); 
				shade(color,shadow,shadowPass,materials,material,currentVal,refletSum,normal,dir,reflectedDir);

				shadowPass++;
			};

			//We have looped over all the lights. Prepare for reflected ray if the surface is reflective.
			//Exit if number of reflection limit was reached or the surface is not reflective.
			if(reflectLimit == 0 || materials[material].reflectivity == 0.0f) break;
			reflectLimit--;

			//The contribution of the reflected ray depends of contribution of previous rays and relfectivity of current material.
			refletSum *= materials[material].reflectivity;

			//Create reflection ray from intersection point and previously computed reflected direction.
			origin = intersectionPoint;
			dir	   = reflectedDir;
		} else
			break;
	};

	//Store the color to buffer as traditional 8-bit RGB.
	{
		unsigned short x = startX + blockIdx.x*8 + threadIdx.x;
		unsigned short y = startY + blockIdx.y*8 + threadIdx.y;
		unsigned index = __mul24(y,width) + x;

		color.x = min(color.x,1.0f);
		color.y = min(color.y,1.0f);
		color.z = min(color.z,1.0f);

		color *= 255.0f;
		uint3 iColor;
		iColor.x = color.x;
		iColor.y = color.y;
		iColor.z = color.z;

		unsigned val = iColor.z<<16 | iColor.y<<8 | iColor.x;
		buffer[index] = val;
	};
};

__global__ void renderGridPS(float3 eye		,float3 upperLeft, float3 dx, float3 dy,unsigned short startX,unsigned short startY,unsigned width,
							 float3 minimum ,float3 maximum, float  cellSize, 
							 uint4* triangles,Material* materials,
							 unsigned int* buffer) {

	//Compute pixel coordinates. The grid is composed of 8x8 blocks.
	unsigned short x  = startX + blockIdx.x*8 + threadIdx.x;
	unsigned short y  = startY + blockIdx.y*8 + threadIdx.y;

	//Compute ray direction for current pixel. To achive right traversal, use higher precision intrinsic.
	float3 dir;
	dir.x = upperLeft.x + __fmul_rn(dx.x,x) + __fmul_rn(dy.x,y);
	dir.y = upperLeft.y + __fmul_rn(dx.y,x) + __fmul_rn(dy.y,y);
	dir.z = upperLeft.z + __fmul_rn(dx.z,x) + __fmul_rn(dy.z,y);

	int hitTriangle = traverseTexture<true>(eye,dir,minimum,maximum,cellSize,triangles);

	unsigned material;
	float3   intersectionPoint;
	float3	 reflectedDir;
	float3   normal;

	if(hitTriangle >=0) {
		float  t;
		
		catchNormalIntersection(triangles,hitTriangle,eye,dir,normal,t,material);

		intersectionPoint = eye + dir*t;						//Compute intersection point of ray and triangle.
		reflectedDir	  = dir - normal*dot(normal,dir)*2.0f;	//Reflected ray computation
	};

	unsigned shadowPass = 0;
	float3   color		= make_float3(0.0f,0.0f,0.0f);	//Start with black pixel color

	if(hitTriangle >= 0)
		while(shadowPass < lightsCount) {
			float3 origin = lights[shadowPass].possition;
			float3 dir	  = intersectionPoint - origin;

			float t = __int_as_float(traverseTexture<false>(origin,dir,minimum,maximum,cellSize,triangles));
	
			bool shadow = t>=0.0f && t<0.999f;
			shade(color,shadow,shadowPass,materials,material,1.0f,1.0f,normal,dir,reflectedDir);

			shadowPass++;
		};

	//Store the color to buffer as traditional 8-bit RGB.
	{
		unsigned short x = startX + blockIdx.x*8 + threadIdx.x;
		unsigned short y = startY + blockIdx.y*8 + threadIdx.y;
		unsigned index = __mul24(y,width) + x;

		color.x = min(color.x,1.0f);
		color.y = min(color.y,1.0f);
		color.z = min(color.z,1.0f);

		color *= 255.0f;
		uint3 iColor;
		iColor.x = color.x;
		iColor.y = color.y;
		iColor.z = color.z;

		unsigned val = iColor.z<<16 | iColor.y<<8 | iColor.x;
		buffer[index] = val;
	};
};

__global__ void renderGridP(float3 eye		,float3 upperLeft, float3 dx, float3 dy,unsigned short startX,unsigned short startY,unsigned width,
						    float3 minimum ,float3 maximum, float  cellSize, 
						    uint4* triangles,Material* materials,
						    unsigned int* buffer) {

	unsigned material;

    float3 dir;
	float3 origin = eye;
	
	//Compute pixel coordinates. The grid is composed of 8x8 blocks.
	unsigned short x  = startX + blockIdx.x*8 + threadIdx.x;
	unsigned short y  = startY + blockIdx.y*8 + threadIdx.y;

	//Compute ray direction for current pixel. To achive right traversal, use higher precision intrinsic.
	dir.x = upperLeft.x + __fmul_rn(dx.x,x) + __fmul_rn(dy.x,y);
	dir.y = upperLeft.y + __fmul_rn(dx.y,x) + __fmul_rn(dy.y,y);
	dir.z = upperLeft.z + __fmul_rn(dx.z,x) + __fmul_rn(dy.z,y);

	//Traverse the grid and try to find intersection.
	int hitTriangle = traverseTexture<true>(origin,dir,minimum,maximum,cellSize,triangles);
	
	float3 color = make_float3(0.0f,0.0f,0.0f);	//Start with black pixel color

	if(hitTriangle >= 0 ) {
		float  t;
		float3 normal;

		//Get intersection parameters for hit triangle. This could be suplied from "traverseTexture" 
		//but is done this way to save some registers. Could be changed to descibed method lates as we 
		//still have some free registers left to next limit and are far away from lowering register usage
		//to get to previous limit.
		catchNormalIntersection(triangles,hitTriangle,eye,dir,normal,t,material);

		float3 intersectionPoint = origin + dir*t;
		float3 reflectedDir = dir - normal*dot(normal,dir)*2.0f;		//Reflected ray computation

		for(int i=0;i<lightsCount;i++) {
			dir = intersectionPoint - lights[i].possition;
			shade(color,false,i,materials,material,1.0f,1.0f,normal,dir,reflectedDir);
		};
	};

	//Store the color to buffer as traditional 8-bit RGB.
	{
		unsigned short x = startX + blockIdx.x*8 + threadIdx.x;
		unsigned short y = startY + blockIdx.y*8 + threadIdx.y;
		unsigned index = __mul24(y,width) + x;

		color.x = min(color.x,1.0f);
		color.y = min(color.y,1.0f);
		color.z = min(color.z,1.0f);

		color *= 255.0f;
		uint3 iColor;
		iColor.x = color.x;
		iColor.y = color.y;
		iColor.z = color.z;

		unsigned val = iColor.z<<16 | iColor.y<<8 | iColor.x;
		buffer[index] = val;
	};
};

float cudaRenderGrid(unsigned int width,unsigned int height,
				    float3 &minimum, float3 &maximum, float cellSize, uint3 &dim,
				    cudaArray* cudaGrid,uint4* triangles,cudaArray* vertices,cudaArray *normals,
				    CameraCuda &camera,unsigned lightsCountIn,Light* lightsIn,Material* materials,
				    unsigned int *buffer,unsigned int *copy,RenderMode mode) {
	if(width % 8 || height % 8) 
		throw 1;

	float3 eye,upperLeft,dx,dy;
	eye = camera.eye;
	upperLeft = camera.upperLeft;
	dx = camera.dx;
	dy = camera.dy;

	float dimX = (float)dim.x;
	float dimY = (float)dim.y;
	float dimZ = (float)dim.z;

	//Prepare out of grid cell coordinates and copy them to constant memory
	float3 prepareOuts[8] = {{-1.0f,-1.0f,-1.0f },{dimX,-1.0f ,-1.0f},{-1.0f,dimY,-1.0f},{dimX,dimY, -1.0f},
						     {-1.0f,-1.0f,dimZ  },{dimX,-1.0f ,dimZ },{-1.0f,dimY,dimZ },{dimX,dimY, dimZ }};
	cudaMemcpyToSymbol(outs,prepareOuts,8*sizeof(int3),0,cudaMemcpyHostToDevice);

	//Prepare grid size limits and copy them to constant memory
	float3 prepareLimits = make_float3(dim.x - 1,dim.y - 1,dim.z - 1);
	cudaMemcpyToSymbol(limits,&prepareLimits,sizeof(float3),0,cudaMemcpyHostToDevice);

	//Copy lights informations to constant memory
	cudaMemcpyToSymbol(lightsCount,&lightsCountIn,sizeof(unsigned)				,0,cudaMemcpyHostToDevice);
	cudaMemcpyToSymbol(lights	  , lightsIn	 ,lightsCountIn*sizeof(Light)   ,0,cudaMemcpyHostToDevice);

	cudaError_t error;

	//Bind the textures with uniform grid structure
	cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<uint2>();
	gridTexture.normalized = false;
	gridTexture.filterMode = cudaFilterModePoint;

	error = cudaBindTextureToArray(gridTexture,cudaGrid,channelDesc);

	channelDesc = cudaCreateChannelDesc<float4>();

	error = cudaBindTextureToArray(vertexTexture,vertices,channelDesc);
	error = cudaBindTextureToArray(normalTexture,normals ,channelDesc);

	//There seemes to be a performace regresion on GTX280 when usin large grid size. The reason is unknown but
	//the solution is to render image as several smaller tiles. Select proper tile size depending on the 
	//image totale resolution.
	
	dim3 grid;
	dim3 threads(8,8,1);
	unsigned startX,startY;
	unsigned tilesX,tilesY;

	if(width<=512) {
		grid.x = width/8;	grid.y = height/8;
		startX = width;		startY = height;
		tilesX = 1;			tilesY = 1;
	} else if(width<768) {
		grid.x = width/16;	grid.y = height/16;
		startX = width/2;	startY = height/2;
		tilesX = 2;			tilesY = 2;
	} else {
		grid.x = width/32;	grid.y = height/32;
		startX = width/4;	startY = height/4;
		tilesX = 4;			tilesY = 4;
	};

	float elapsedTime;

	cudaEvent_t start,stop;
	cudaEventCreate(&start);
	cudaEventCreate(&stop);

	cudaEventRecord(start,0);


	switch(mode) {
		case PRIMARY:
			for(unsigned y=0;y<tilesX;y++)
				for(unsigned x=0;x<tilesY;x++) 
					renderGridP  <<<grid,threads>>>(eye,upperLeft,dx,dy,startX*x,startY*y,width,minimum,maximum,cellSize,triangles,materials,buffer);
			break;
		case PRIMARY_SHADOW:
			for(unsigned y=0;y<tilesX;y++)
				for(unsigned x=0;x<tilesY;x++) 
					renderGridPS <<<grid,threads>>>(eye,upperLeft,dx,dy,startX*x,startY*y,width,minimum,maximum,cellSize,triangles,materials,buffer);
			break;
		case PRIMARY_SHADOW_REFLECT:
			for(unsigned y=0;y<tilesX;y++)
				for(unsigned x=0;x<tilesY;x++) 
					renderGridPSR<<<grid,threads>>>(eye,upperLeft,dx,dy,startX*x,startY*y,width,minimum,maximum,cellSize,triangles,materials,buffer);
			break;
	};

	cudaEventRecord(stop,0);
	cudaEventSynchronize(stop);

	cudaEventElapsedTime(&elapsedTime,start,stop);

	cudaEventDestroy(start);
	cudaEventDestroy(stop);

	cudaUnbindTexture(gridTexture);
	cudaUnbindTexture(vertexTexture);
	cudaUnbindTexture(normalTexture);

	//Copy the image buffer to main memory.
	cudaMemcpy(copy,buffer,width*height*4,cudaMemcpyDeviceToHost);

	return elapsedTime;
};

//Same as above, just using OpenGL PBO instead of explicitly alocated buffer
float cudaRenderGridTexture(unsigned int width,unsigned int height,
				    float3 &minimum, float3 &maximum, float cellSize, uint3 &dim,
				    cudaArray* cudaGrid,uint4* triangles,cudaArray* vertices,cudaArray *normals,
				    CameraCuda &camera,unsigned lightsCountIn,Light* lightsIn,Material* materials,
				    unsigned int pbo,RenderMode mode) {
	if(width % 8 || height % 8) 
		throw 1;

	float3 eye,upperLeft,dx,dy;
	eye = camera.eye;
	upperLeft = camera.upperLeft;
	dx = camera.dx;
	dy = camera.dy;

	float dimX = (float)dim.x;
	float dimY = (float)dim.y;
	float dimZ = (float)dim.z;

	float3 prepareOuts[8] = {{-1.0f,-1.0f,-1.0f },{dimX,-1.0f ,-1.0f},{-1.0f,dimY,-1.0f},{dimX,dimY, -1.0f},
						     {-1.0f,-1.0f,dimZ  },{dimX,-1.0f ,dimZ },{-1.0f,dimY,dimZ },{dimX,dimY, dimZ }};
	cudaMemcpyToSymbol(outs,prepareOuts,8*sizeof(int3),0,cudaMemcpyHostToDevice);

	float3 prepareLimits = make_float3(dim.x - 1,dim.y - 1,dim.z - 1);
	cudaMemcpyToSymbol(limits,&prepareLimits,sizeof(float3),0,cudaMemcpyHostToDevice);

	cudaMemcpyToSymbol(lightsCount,&lightsCountIn,sizeof(unsigned)				,0,cudaMemcpyHostToDevice);
	cudaMemcpyToSymbol(lights	  , lightsIn	 ,lightsCountIn*sizeof(Light)   ,0,cudaMemcpyHostToDevice);

	cudaError_t error;

	cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<uint2>();
	gridTexture.normalized = false;
	gridTexture.filterMode = cudaFilterModePoint;

	error = cudaBindTextureToArray(gridTexture,cudaGrid,channelDesc);

	channelDesc = cudaCreateChannelDesc<float4>();

	error = cudaBindTextureToArray(vertexTexture,vertices,channelDesc);
	error = cudaBindTextureToArray(normalTexture,normals ,channelDesc);

	unsigned int *buffer;
	error = cudaGLMapBufferObject((void**)&buffer,pbo);

	dim3 grid;
	dim3 threads(8,8,1);
	unsigned startX,startY;
	unsigned tilesX,tilesY;

	if(width<=512) {
		grid.x = width/8;	grid.y = height/8;
		startX = width;		startY = height;
		tilesX = 1;			tilesY = 1;
	} else if(width<768) {
		grid.x = width/16;	grid.y = height/16;
		startX = width/2;	startY = height/2;
		tilesX = 2;			tilesY = 2;
	} else {
		grid.x = width/32;	grid.y = height/32;
		startX = width/4;	startY = height/4;
		tilesX = 4;			tilesY = 4;
	};

	float elapsedTime;

	cudaEvent_t start,stop;
	cudaEventCreate(&start);
	cudaEventCreate(&stop);

	cudaEventRecord(start,0);

	switch(mode) {
		case PRIMARY:
			for(unsigned y=0;y<tilesX;y++)
				for(unsigned x=0;x<tilesY;x++) 
					renderGridP  <<<grid,threads>>>(eye,upperLeft,dx,dy,startX*x,startY*y,width,minimum,maximum,cellSize,triangles,materials,buffer);
			break;
		case PRIMARY_SHADOW:
			for(unsigned y=0;y<tilesX;y++)
				for(unsigned x=0;x<tilesY;x++) 
					renderGridPS <<<grid,threads>>>(eye,upperLeft,dx,dy,startX*x,startY*y,width,minimum,maximum,cellSize,triangles,materials,buffer);
			break;
		case PRIMARY_SHADOW_REFLECT:
			for(unsigned y=0;y<tilesX;y++)
				for(unsigned x=0;x<tilesY;x++) 
					renderGridPSR<<<grid,threads>>>(eye,upperLeft,dx,dy,startX*x,startY*y,width,minimum,maximum,cellSize,triangles,materials,buffer);
			break;
	};

	cudaEventRecord(stop,0);
	cudaEventSynchronize(stop);

	cudaEventElapsedTime(&elapsedTime,start,stop);

	cudaEventDestroy(start);
	cudaEventDestroy(stop);

	cudaGLUnmapBufferObject(pbo);

	cudaUnbindTexture(gridTexture);
	cudaUnbindTexture(vertexTexture);
	cudaUnbindTexture(normalTexture);

	return elapsedTime;
};
