/*
**	  GPU Raytrace - Diploma thesis, Czech technical university
**	  Copyright (C) 2008 Martin Zlatuska
**
**	  This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <float.h>

#include "Utils/GLEE/glew.h"
#include <cuda_gl_interop.h>

#include "vectorMath.cu"
#include "RenderMode.h"

#include "common.cu"

texture<uint2 , 2, cudaReadModeElementType> cellTexture;	//Texture used for storing tree nodes.

const unsigned LEAF = 0x03;


//! Traverse kd tree with ray.
/**
*	@param[in] eye			Position of ray origin.
*	@param[in] dir			Ray direction.
*	@param[in] minimum		Minimum extent of the root node bounding box.
*	@param[in] maximum		Maximum extent of the root node bounding box.
*	@param[in] triangles	
*	@param[in] buffers
*	@param[in] stackNode	Stack array for node number.
*	@param[in] stackMin		Stack array for storing ray minimum of active extent.
*	@param[in] stackMax		Stack array for storing ray maximum of active extent.
*/
template<bool output>
__device__ int traverseKdTree(float3 &eye,float3 &dir,float3 &minimum ,float3 &maximum, uint4* triangles,unsigned* buffer,
							  unsigned stackNode[272],float	stackMin[272],float stackMax[272]) {
	float tMin,tMax;
	float globalMax;
	
	int stackStart = 0;
	int stackCount = 0;
	int index	   = (threadIdx.y*8 + threadIdx.x)*4;	//Index to the begining of this ray stack.
	index		  += (threadIdx.y*8 + threadIdx.x)>>2;	//To avoid bank conflict, pad the index every 4 threads.

	unsigned startNode = 0;
	bool	 pushDown  = true;

	// Kay and Kayjia ray-box intersection test as descibed in "An introduction to Ray Tracing" by A.S.Glassner
	intersectBox(minimum,maximum,eye,dir,tMin,tMax);
	tMin = max(tMin,0.0f);
	globalMax = tMax; 

	int   hitTriangle = -1;
	float closestHit  = FLT_MAX;

	unsigned node = 0;

	while(hitTriangle==-1) {	
		//Exit if the ray leaves the tree without a hit. This also exit if ray miss the initial bounding box.
		if(tMin >= tMax) break;
		if(!output && tMin>=1.0f) break;	//If testing shadow ray, exit if possible inersection is farther 
											//than the original intersection, and therefore cannot possibly cast shadow.

		//The texture must be 2D because of the 1D size limit.
		unsigned nodeX = node &  255;
		unsigned nodeY = node >> 8;

		uint2 cell = tex2D(cellTexture,nodeX,nodeY);	//Catch cell info from texture

		unsigned type = cell.x & 0x3;
		if(type == LEAF) {
			unsigned count  = cell.x >> 2;
			unsigned offset = cell.y;

			//Intersect the ray with this leaf's triangles.
			for(int i=0;i<count;i++) {
				float u,v,t;
				intersectTexture(triangles,eye,dir,offset,u,v,t);

				if(output) {
					if(v>=-0.00048828125f && u>=-0.00048828125f && u+v<=1.0000304f && t>=0.010070801f && t<=closestHit && t<=tMax+0.00048828125f) {// && t>=tMin-0.48828125f) {
						closestHit  = t;
						hitTriangle = offset; 
					};
				} else {
					//If testing shadow ray, accept the intersection even if outside the node.
					if(v>=-0.00048828125f && u>=-0.00048828125f && u+v<=1.0000304f && t>=0.010070801f && t<=0.999f) {
						closestHit  = t;
						hitTriangle = offset; 
					};
				};

				offset++;
			};

			//Exit as soon as intersection is found.
			if(hitTriangle!=-1) break;

			// If we miss and have node on stack, pop it.
			if(stackCount) {
				stackCount--;
				stackStart = (stackStart-1)&0x3;

				node = stackNode[index + stackStart];
				tMin = stackMin [index + stackStart];
				tMax = stackMax [index + stackStart];
			} else {	//Otherwise we must restart from known restart point.
				tMin = tMax;
				tMax = globalMax;
				node = startNode;
				pushDown = true;
			};
		} else {
			float split = __int_as_float(cell.y);

			float eyeDim;
			float dirDim;

			//Gather data for split plane intersection. Unfortunately, as the data are in registers, they 
			//cannot be accesed using index. One possibility is to copy to shared memory and index there.
			//But that would consumes to much memory.
			switch(type) {
				case 0: eyeDim = eye.x; dirDim = dir.x; break;
				case 1: eyeDim = eye.y; dirDim = dir.y; break;
				case 2: eyeDim = eye.z; dirDim = dir.z; break;
			};

			float tHit   = (split - eyeDim)/dirDim;
			
			//Find witch cell child is entered first by using the split/eye possition
			unsigned first,second;
			unsigned left  = node + 1;
			unsigned right = cell.x >> 2;

			//Try to correctly find whitch of the child nodes is near and far. This has many branches but the absolute majority
			//of the rays will take the first or second branch so it is not really that painfull.
			if(eyeDim < split) {
				first   = left;
				second  = right;
			} else if(eyeDim > split) {
				first   = right;
				second  = left;
			} else if(dirDim < 0.0f) {
				first   = right;
				second  = left;
			} else {
				first   = left;
				second  = right;
			}

			//If the ray is coplanar to the spliting plane, the ray active range doesn't change and tHit is invalid, so set it to current maximum.
			if(dirDim == 0.0f) tHit = tMax;

			//Now find, which of the child nodes are intersected by ray.
			//The need for testing for equality because of the restart unfortunately causes wrong hadnling of zero volume cells.
			if(tHit >= tMax || tHit<0) {	//Only the first child is intersected
				startNode = pushDown ? first : startNode;
				node = first;	
			} else if(tHit <= tMin ) {		//Only the second child is intersected
				startNode = pushDown ? second : startNode;
				node = second;
			} else {//Both child need to be traversed. Push second on the stack and move to first.
				pushDown = false; //The push-down of restart node must end at this point.

				node = first;

				//Push the farther node on stack.
				stackNode[index + stackStart] = second;
				stackMin [index + stackStart] = tHit;
				stackMax [index + stackStart] = tMax;

				//The stack is really quite small and will probably often overflow. In that case tha last item on
				//stack is discarded. To do this, the current stack top indes is simply looped around to the start of the
				//stack array, similar to circular buffer.
				stackStart = (stackStart+1) & 0x3;	
				stackCount = min(stackCount+1,4);	//Increase stack count and clamp to maximum stack size.

				tMax = tHit;
			};
		};
	};
	
	if(output)
		return hitTriangle;
	else
		return hitTriangle == -1 ? __float_as_int(-1.0f) :  __float_as_int(closestHit);
};

__global__ void renderTreePSR(float3 eye     ,float3 upperLeft	, float3 dx, float3 dy,unsigned short startX,unsigned short startY,unsigned width,
							  float3 minimum ,float3 maximum	,
							  uint4* triangles,Material* materials,
							  unsigned int* buffer) {
	//Define the arrays for stack here. The current compiler version allocate shared memory for every called 
	//function and it seams it remains reserved during whole kernel run. Because of the use of template, the traversal 
	//fuction is trated as two distict functions and if declared inside the funciton, the use of shared memory doubles.
	__shared__ unsigned stackNode[272];
	__shared__ float	stackMin[272];
	__shared__ float	stackMax[272];

	unsigned	reflectLimit = 2;		//Limit number of reflection to fixed number. Might change to paramtere later
	float		refletSum	 = 1.0f;	

	//Compute pixel coordinates. The grid is composed of 8x8 blocks.
	unsigned short x  = startX + blockIdx.x*8 + threadIdx.x;
	unsigned short y  = startY + blockIdx.y*8 + threadIdx.y;
	
	//Compute ray direction for current pixel. To achive right traversal, use higher precision intrinsic.
	float3 dir;
	dir.x = upperLeft.x + __fmul_rn(dx.x,x) + __fmul_rn(dy.x,y);
	dir.y = upperLeft.y + __fmul_rn(dx.y,x) + __fmul_rn(dy.y,y);
	dir.z = upperLeft.z + __fmul_rn(dx.z,x) + __fmul_rn(dy.z,y);

	float3 color = make_float3(0.0f,0.0f,0.0f);	//Start with black pixel color

	//if(!(blockIdx.x == 50 && blockIdx.y == 30))	return;

	float3 origin = eye;
	while(true) {
		int hitTriangle = traverseKdTree<true>(origin,dir,minimum,maximum,triangles,buffer,stackNode,stackMin,stackMax);

		unsigned material;
		float3   intersectionPoint;
		float3	 reflectedDir;
		float3   normal;

		if(hitTriangle >=0) {
			float  t;
			
			catchNormalIntersection(triangles,hitTriangle,origin,dir,normal,t,material);

			intersectionPoint = origin + dir*t;							//Compute intersection point of ray and triangle.
			reflectedDir	  = dir    - normal*dot(normal,dir)*2.0f;	//Reflected ray computation
		};

		unsigned shadowPass = 0;

		if(hitTriangle >= 0) {
			while(shadowPass < lightsCount) {
				origin = lights[shadowPass].possition;
				dir    = intersectionPoint - origin;

				float t = __int_as_float(traverseKdTree<false>(origin,dir,minimum,maximum,triangles,buffer,stackNode,stackMin,stackMax));
		
				bool shadow = t>=0.0f && t<0.999f;
				float currentVal = refletSum*(1.0f - materials[material].reflectivity); 
				shade(color,shadow,shadowPass,materials,material,currentVal,refletSum,normal,dir,reflectedDir);

				shadowPass++;
			};

			//We have looped over all the lights. Prepare for reflected ray if the surface is reflective.
			//Exit if number of reflection limit was reached or the surface is not reflective.
			if(reflectLimit == 0 || materials[material].reflectivity == 0.0f) break;
			reflectLimit--;

			//The contribution of the reflected ray depends of contribution of previous rays and relfectivity of current material.
			refletSum *= materials[material].reflectivity;

			//Create reflection ray from intersection point and previously computed reflected direction.
			origin = intersectionPoint;
			dir	   = reflectedDir;
		} else
			break;
	};

	//Store the color to buffer as traditional 8-bit RGB.
	storeInteger(color,buffer,startX,startY,width);
};

__global__ void renderTreePS(float3 eye		,float3 upperLeft, float3 dx, float3 dy,unsigned short startX,unsigned short startY,unsigned width,
							 float3 minimum ,float3 maximum, 
							 uint4* triangles,Material* materials,
							 unsigned int* buffer) {
	__shared__ unsigned stackNode[272];
	__shared__ float	stackMin[272];
	__shared__ float	stackMax[272];

	//Compute pixel coordinates. The grid is composed of 8x8 blocks.
	unsigned short x  = startX + blockIdx.x*8 + threadIdx.x;
	unsigned short y  = startY + blockIdx.y*8 + threadIdx.y;

	//Compute ray direction for current pixel. To achive right traversal, use higher precision intrinsic.
	float3 dir;
	dir.x = upperLeft.x + __fmul_rn(dx.x,x) + __fmul_rn(dy.x,y);
	dir.y = upperLeft.y + __fmul_rn(dx.y,x) + __fmul_rn(dy.y,y);
	dir.z = upperLeft.z + __fmul_rn(dx.z,x) + __fmul_rn(dy.z,y);

	int hitTriangle = traverseKdTree<true>(eye,dir,minimum,maximum,triangles,buffer,stackNode,stackMin,stackMax);

	unsigned material;
	float3   intersectionPoint;
	float3	 reflectedDir;
	float3   normal;

	if(hitTriangle >=0) {
		float  t;
		
		catchNormalIntersection(triangles,hitTriangle,eye,dir,normal,t,material);

		intersectionPoint = eye + dir*t;						//Compute intersection point of ray and triangle.
		reflectedDir	  = dir - normal*dot(normal,dir)*2.0f;	//Reflected ray computation
	};

	unsigned shadowPass = 0;
	float3   color		= make_float3(0.0f,0.0f,0.0f);	//Start with black pixel color

	if(hitTriangle >= 0)
		while(shadowPass < lightsCount) {
			float3 origin = lights[shadowPass].possition;
			float3 dir	= intersectionPoint - origin;

			float t = __int_as_float(traverseKdTree<false>(origin,dir,minimum,maximum,triangles,buffer,stackNode,stackMin,stackMax));
	
			bool shadow = t>=0.0f && t<0.999f;
			shade(color,shadow,shadowPass,materials,material,1.0f,1.0f,normal,dir,reflectedDir);

			shadowPass++;
		};

	//Store the color to buffer as traditional 8-bit RGB.
	storeInteger(color,buffer,startX,startY,width);
};

__global__ void renderTreeP(float3 eye	   ,float3 upperLeft, float3 dx, float3 dy,unsigned short startX,unsigned short startY,unsigned width,
						    float3 minimum ,float3 maximum,
						    uint4* triangles,Material* materials,
						    unsigned int* buffer) {
	__shared__ unsigned stackNode[272];
	__shared__ float	stackMin[272];
	__shared__ float	stackMax[272];

	unsigned material;

    float3 dir;
	float3 origin = eye;
	
	//Compute pixel coordinates. The grid is composed of 8x8 blocks.

	unsigned short x  = startX + blockIdx.x*8 + threadIdx.x;
	unsigned short y  = startY + blockIdx.y*8 + threadIdx.y;
		
	//Compute ray direction for current pixel. To achive right traversal, use higher precision intrinsic.
	dir.x = upperLeft.x + __fmul_rn(dx.x,x) + __fmul_rn(dy.x,y);
	dir.y = upperLeft.y + __fmul_rn(dx.y,x) + __fmul_rn(dy.y,y);
	dir.z = upperLeft.z + __fmul_rn(dx.z,x) + __fmul_rn(dy.z,y);

	//if(!(dir.y == 0.0f || dir.x==0.0f || dir.z==0.0f))	return;

	//Traverse the grid and try to find intersection.
	int hitTriangle = traverseKdTree<true>(origin,dir,minimum,maximum,triangles,buffer,stackNode,stackMin,stackMax);

	float3 color = make_float3(0.0f,0.0f,0.0f);	//Start with black pixel color

	if(hitTriangle >= 0) {
		float  t;
		float3 normal;

		//Get intersection parameters for hit triangle. This could be suplied from "traverseTexture" 
		//but is done this way to save some registers. Could be changed to descibed method lates as we 
		//still have some free registers left to next limit and are far away from lowering register usage
		//to get to previous limit.
		catchNormalIntersection(triangles,hitTriangle,eye,dir,normal,t,material);

		float3 intersectionPoint = origin + dir*t;
		float3 reflectedDir = dir - normal*dot(normal,dir)*2.0f;		//Reflected ray computation

		for(int i=0;i<lightsCount;i++) {
			dir = intersectionPoint - lights[i].possition;
			shade(color,false,i,materials,material,1.0f,1.0f,normal,dir,reflectedDir);
		};
	};

	//Store the color to buffer as traditional 8-bit RGB.
	storeInteger(color,buffer,startX,startY,width);
};

float cudaRenderKd(unsigned int width,unsigned int height,
				    float3 &minimum, float3 &maximum,
				    cudaArray* cudaTree,uint4* triangles,cudaArray* vertices,cudaArray *normals,
				    CameraCuda &camera,unsigned lightsCountIn,Light* lightsIn,Material* materials,
				    unsigned int *buffer,unsigned int *copy,RenderMode mode) {
	if(width % 8 || height % 8) 
		throw 1;

	float3 eye,upperLeft,dx,dy;
	eye = camera.eye;
	upperLeft = camera.upperLeft;
	dx = camera.dx;
	dy = camera.dy;


	cudaMemcpyToSymbol(lightsCount,&lightsCountIn,sizeof(unsigned)				,0,cudaMemcpyHostToDevice);
	cudaMemcpyToSymbol(lights	  , lightsIn	 ,lightsCountIn*sizeof(Light)   ,0,cudaMemcpyHostToDevice);

	cudaError_t error;

	cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<uint2>();
	cellTexture.normalized = false;
	cellTexture.filterMode = cudaFilterModePoint;
	error = cudaBindTextureToArray(cellTexture,cudaTree,channelDesc);

	channelDesc = cudaCreateChannelDesc<float4>();
	error = cudaBindTextureToArray(vertexTexture,vertices,channelDesc);
	error = cudaBindTextureToArray(normalTexture,normals ,channelDesc);

	dim3 grid;
	dim3 threads(8,8,1);
	unsigned startX,startY;
	unsigned tilesX,tilesY;

	if(width<=512) {
		grid.x = width/8;	grid.y = height/8;
		startX = width;		startY = height;
		tilesX = 1;			tilesY = 1;
	} else if(width<768) {
		grid.x = width/16;	grid.y = height/16;
		startX = width/2;	startY = height/2;
		tilesX = 2;			tilesY = 2;
	} else {
		grid.x = width/32;	grid.y = height/32;
		startX = width/4;	startY = height/4;
		tilesX = 4;			tilesY = 4;
	};


	float elapsedTime;

	cudaEvent_t start,stop;
	cudaEventCreate(&start);
	cudaEventCreate(&stop);

	cudaEventRecord(start,0);

	switch(mode) {
		case PRIMARY:
			for(unsigned y=0;y<tilesX;y++)
				for(unsigned x=0;x<tilesY;x++) 
					renderTreeP  <<<grid,threads>>>(eye,upperLeft,dx,dy,startX*x,startY*y,width,minimum,maximum,triangles,materials,buffer);
			break;
		case PRIMARY_SHADOW:
			for(unsigned y=0;y<tilesX;y++)
				for(unsigned x=0;x<tilesY;x++) 
					renderTreePS <<<grid,threads>>>(eye,upperLeft,dx,dy,startX*x,startY*y,width,minimum,maximum,triangles,materials,buffer);
			break;
		case PRIMARY_SHADOW_REFLECT:
			for(unsigned y=0;y<tilesX;y++)
				for(unsigned x=0;x<tilesY;x++) 
					renderTreePSR<<<grid,threads>>>(eye,upperLeft,dx,dy,startX*x,startY*y,width,minimum,maximum,triangles,materials,buffer);
			break;
	};

	cudaEventRecord(stop,0);
	cudaEventSynchronize(stop);

	cudaEventElapsedTime(&elapsedTime,start,stop);

	cudaEventDestroy(start);
	cudaEventDestroy(stop);

	cudaUnbindTexture(cellTexture);
	cudaUnbindTexture(vertexTexture);
	cudaUnbindTexture(normalTexture);

	cudaMemcpy(copy,buffer,width*height*4,cudaMemcpyDeviceToHost);

	return elapsedTime;
};

float cudaRenderKdTexture(unsigned int width,unsigned int height,
				    float3 &minimum, float3 &maximum,
				    cudaArray* cudaTree,uint4* triangles,cudaArray* vertices,cudaArray *normals,
				    CameraCuda &camera,unsigned lightsCountIn,Light* lightsIn,Material* materials,
				    unsigned int pbo,RenderMode mode) {
	if(width % 8 || height % 8) 
		throw 1;

	float3 eye,upperLeft,dx,dy;
	eye = camera.eye;
	upperLeft = camera.upperLeft;
	dx = camera.dx;
	dy = camera.dy;

	cudaMemcpyToSymbol(lightsCount,&lightsCountIn,sizeof(unsigned)				,0,cudaMemcpyHostToDevice);
	cudaMemcpyToSymbol(lights	  , lightsIn	 ,lightsCountIn*sizeof(Light)   ,0,cudaMemcpyHostToDevice);

	cudaError_t error;

	cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<uint2>();
	cellTexture.normalized = false;
	cellTexture.filterMode = cudaFilterModePoint;
	error = cudaBindTextureToArray(cellTexture,cudaTree,channelDesc);

	channelDesc = cudaCreateChannelDesc<float4>();
	error = cudaBindTextureToArray(vertexTexture,vertices,channelDesc);
	error = cudaBindTextureToArray(normalTexture,normals ,channelDesc);

	dim3 grid;
	dim3 threads(8,8,1);
	unsigned startX,startY;
	unsigned tilesX,tilesY;

	if(width<=512) {
		grid.x = width/8;	grid.y = height/8;
		startX = width;		startY = height;
		tilesX = 1;			tilesY = 1;
	} else if(width<768) {
		grid.x = width/16;	grid.y = height/16;
		startX = width/2;	startY = height/2;
		tilesX = 2;			tilesY = 2;
	} else {
		grid.x = width/32;	grid.y = height/32;
		startX = width/4;	startY = height/4;
		tilesX = 4;			tilesY = 4;
	};

	unsigned int *buffer;
	error = cudaGLMapBufferObject((void**)&buffer,pbo);

	float elapsedTime;

	cudaEvent_t start,stop;
	cudaEventCreate(&start);
	cudaEventCreate(&stop);

	cudaEventRecord(start,0);

	switch(mode) {
		case PRIMARY:
			for(unsigned y=0;y<tilesX;y++)
				for(unsigned x=0;x<tilesY;x++) 
					renderTreeP  <<<grid,threads>>>(eye,upperLeft,dx,dy,startX*x,startY*y,width,minimum,maximum,triangles,materials,buffer);
			break;
		case PRIMARY_SHADOW:
			for(unsigned y=0;y<tilesX;y++)
				for(unsigned x=0;x<tilesY;x++) 
					renderTreePS <<<grid,threads>>>(eye,upperLeft,dx,dy,startX*x,startY*y,width,minimum,maximum,triangles,materials,buffer);
			break;
		case PRIMARY_SHADOW_REFLECT:
			for(unsigned y=0;y<tilesX;y++)
				for(unsigned x=0;x<tilesY;x++) 
					renderTreePSR<<<grid,threads>>>(eye,upperLeft,dx,dy,startX*x,startY*y,width,minimum,maximum,triangles,materials,buffer);
			break;
	};

	cudaEventRecord(stop,0);
	cudaEventSynchronize(stop);

	cudaEventElapsedTime(&elapsedTime,start,stop);

	cudaEventDestroy(start);
	cudaEventDestroy(stop);

	cudaGLUnmapBufferObject(pbo);

	cudaUnbindTexture(cellTexture);
	cudaUnbindTexture(vertexTexture);
	cudaUnbindTexture(normalTexture);

	return elapsedTime;
};

