/*
**	  GPU Raytrace - Diploma thesis, Czech technical university
**	  Copyright (C) 2008 Martin Zlatuska
**
**	  This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <emmintrin.h>

#include <channel_descriptor.h>
#include <driver_functions.h>
#include <cuda_runtime_api.h>
#include <cuda.h>

#include "Material.h"
#include "Utils/Memory.h"

#include "UniformGrid.h"

#include "time.h"

#ifdef _WIN32 
#include <windows.h>
#endif

using std::max;
using std::min;

using sse::operator+;
using sse::operator+=;
using sse::operator-;
using sse::operator*;
using sse::operator/;
using sse::operator<;
using sse::operator==;
using sse::operator|;
using sse::operator&;

#undef min
#undef max

//#define MEASURE_TIME

struct TriangleExtent {
	signed minX,minY,minZ;
	signed maxX,maxY,maxZ;
};

UniformGrid::UniformGrid(Scene &scene,int minimalSideSize) {
	videoAllocated = false;
	int vertexSum = 0;
	int facesSum  = 0;

	/////////////////////////////
	// Find grid extents. The individual object's bounding boxes are transformed
	// into world space and the minimum and maximum is found;
	/////////////////////////////

	//Initialize minimum and maximum to oposite limits to ensure first comparation results in assigning the value
	__m128 minimum,maximum;
	minimum = minimumInit;
	maximum = maximumInit;

	size_t sceneSize = scene.size();
	for(size_t i=0;i<sceneSize;i++) {
		//Prepare total number of faces and vertieces in grid
		vertexSum += scene[i].object->vertexCount;
		facesSum  += scene[i].object->facesCount;

		Vertex &min = scene[i].object->minimum;
		Vertex &max = scene[i].object->maximum;

		//Transform all corners of bounding box into world space and then check for minimality/maximality
		for(int j=0;j<8;j++) {
			SPVector v1;
			v1.x = j & 0x01 ? min.x : max.x;
			v1.y = j & 0x02 ? min.y : max.y;
			v1.z = j & 0x04 ? min.z : max.z;
			v1.w = 1.0f;

			v1 *= *scene[i].transformation;

			minimum = sse::min(minimum,v1.vec);
			maximum = sse::max(maximum,v1.vec);
		};
	};

	#ifdef MEASURE_TIME
		#ifdef _WIN32
			__int64 ticks,start,stop;
			QueryPerformanceFrequency((LARGE_INTEGER*)&ticks);
			double TICKS_PER_SEC=ticks;
			QueryPerformanceCounter((LARGE_INTEGER*)&start);
		#else
			clock_t start,stop;
			start = clock();
		#endif
	#endif

	ALIGN16 float fMaximum[4];
	ALIGN16 float fMinimum[4];
			
	//Store grid origin for cuda
	sse::store(minimum,fMinimum);
	cuMinimum.x = fMinimum[0];
	cuMinimum.y = fMinimum[1];
	cuMinimum.z = fMinimum[2];

	//Compute size of cell in each dimension for minimal requested cell size
	maximum = maximum - minimum;

	if(minimalSideSize==-1) {
		minimalSideSize = sqrtf(sqrtf(facesSum))*4;
		minimalSideSize = min(minimalSideSize,128);
	};

	sse::store(maximum,fMaximum);
	float x = fMaximum[0] / minimalSideSize;
	float y = fMaximum[1] / minimalSideSize;
	float z = fMaximum[2] / minimalSideSize;

	//Find minimal dimension
	float minSize = min(min(x,y),z);
	cuCellSize = minSize;

	//Cell has the same size in every dimension. Now find the number of cells in each dimension for cell size
	dimX = (int)ceilf(fMaximum[0] / minSize);
	dimY = (int)ceilf(fMaximum[1] / minSize);
	dimZ = (int)ceilf(fMaximum[2] / minSize);

	//As the number of cells is integer, the maximum extend of the grid is now not the original. Recompute.
	cuMaximum.x = cuMinimum.x + dimX*minSize;
	cuMaximum.y = cuMinimum.y + dimY*minSize;
	cuMaximum.z = cuMinimum.z + dimZ*minSize;

	SPVector origin = minimum;

	SPVector cellSize;
	cellSize.x = cellSize.y = cellSize.z = 1.0f/minSize;

	/////////////////////////////////////
	// First pass
	// Transform object vertives into world space, and find the cells into each triangle belongs
	/////////////////////////////////////

	//Allocate data
	grid			= new Cell[dimX*dimY*dimZ];
	transformed		= new Vertex[vertexSum];
	normals			= new Vertex[vertexSum];

	totalVertex = vertexSum;

	memset(grid,0,dimX*dimY*dimZ*sizeof(Cell));

	int curVertex = 0;
	int curFace   = 0;


	//We need compact storage of the grid's triangle indexes in one array. 
	//As we do not know how many triangles are there in grid cell, we must do in in a two passes.
	//First we transform triangle in world space, found its extend in grid and increase the counters in respective cells.
	//To save computation, we remember the triangle's extents for second pass.
	//Then the triangle count is sumed producing offset for each cell from begining of the array.
	//Now we can allocate data and add pointers to faces into grid.

	TriangleExtent* __restrict extents = new TriangleExtent[facesSum];

	for(unsigned i=0;i<sceneSize;i++) {

		int facesCount  = scene[i].object->facesCount;
		int vertexCount = scene[i].object->vertexCount;


		Vertex*			__restrict vertices	 = scene[i].object->vertices;
		Vertex*			__restrict normalsIn = scene[i].object->normal;
		Triangle*		__restrict faces	 = scene[i].object->faces;
		Transformation* __restrict matrix	 = scene[i].transformation;
		Vertex*			__restrict transformedRebased = &transformed[curVertex];

		//Transform all vertices and add them to the grid.
		for(int j=0;j<vertexCount;j++) {
			transformed[curVertex].vec = ( vertices[j] * *matrix).vec;
			normals[curVertex++]	   = normalsIn[j];
		};

		//Set the SSE rounding to round to zero. We need this for triangle extent conversion to integer.
		_MM_SET_ROUNDING_MODE(_MM_ROUND_TOWARD_ZERO);

		ALIGN16 int4 iMin;
		ALIGN16 int4 iMax;

		for(int j=0;j<facesCount;j++) {
			//Convert vertices coordinates to grid index;
			SPVector v1 = (transformedRebased[faces[j].a] - origin) | cellSize; 
			SPVector v2 = (transformedRebased[faces[j].b] - origin) | cellSize; 
			SPVector v3 = (transformedRebased[faces[j].c] - origin) | cellSize; 

			//Find extents
			__m128i minimum,maximum;
			minimum = sse::f4i(sse::min(sse::min(v2.vec,v3.vec),v1.vec));
			maximum = sse::f4i(sse::max(sse::max(v2.vec,v3.vec),v1.vec));

			//Load from SSE
			sse::store(minimum,&iMin);
			sse::store(maximum,&iMax);
			
			iMax.x = min(iMax.x,dimX-1);
			iMax.y = min(iMax.y,dimY-1);
			iMax.z = min(iMax.z,dimZ-1);			

			//Save extents for second pass
			extents[curFace].minX = iMin.x;
			extents[curFace].minY = iMin.y;
			extents[curFace].minZ = iMin.z;
			extents[curFace].maxX = iMax.x;
			extents[curFace].maxY = iMax.y;
			extents[curFace].maxZ = iMax.z;
			curFace++;

			//Increase counter in triangle's cells
			for(int z=iMin.z;z<=iMax.z;z++) 
				for(int y=iMin.y;y<=iMax.y;y++) 
					for(int x=iMin.x;x<=iMax.x;x++)
						grid[z*dimX*dimY + y*dimX + x].count++;
		};
	};

	/////////////////////////////////////
	// Second pass.
	// Use cell triangle count from the first pass to compute the possition of the cell triangle references.
	// Than we can have all triangle references in one large array.
	/////////////////////////////////////

	//Sum total number of used faces and cell offset into array;
	unsigned sum = 0;
	for(int i=0;i<dimX*dimY*dimZ;i++) {
		grid[i].offset = sum;
		sum += grid[i].count;
	};

	//Now we can allocate memory for triangle array.
	totalPoly	  = sum;
	cellTriangles = new Triangle[sum];
	
	int object = 0;
	int index  = 0;
	curFace	   = 0;

	//Allocate and clean the cell triangle usage.
	unsigned* offsets = new unsigned[dimX*dimY*dimZ];
	memset(offsets,0,dimX*dimY*dimZ*4);

	vertexSum = 0;
	__m128i vertexSumSse = sse::set(0,0,0,0);
	__m128i maskFirst    = sse::set(0,0xFFFFFFFF,0xFFFFFFFF,0xFFFFFFFF);

	//Now we have the prefix sum of cells offset and can save the faces references in one large linear array.
	for(unsigned i=0;i<sceneSize;i++) {
		Object *object = scene[i].object;
		int size = object->facesCount;

		//Check for replacement materials in the object instance in the scene. Set the correct material source.
		std::vector<unsigned> &materialIndexes = scene[i].replacementMaterials.size() ? scene[i].replacementMaterials : object->materials;

		//For every face in object.
		for(int j=0;j<size;j++) {
			__m128i triangle = sse::load<__m128i>(&object->faces[j]);				//Load the triangle
			__m128i material = sse::pack(materialIndexes[sse::extract(triangle)]);	//Change the object material number to global material number.

			//Mask the material into the triangle
			triangle = triangle & maskFirst;
			triangle = triangle | material;

			//Convert object triangle indices to global indices
			triangle += vertexSumSse;

			for(int z=extents[curFace].minZ;z<=extents[curFace].maxZ;z++) 
				for(int y=extents[curFace].minY;y<=extents[curFace].maxY;y++) 
					for(int x=extents[curFace].minX;x<=extents[curFace].maxX;x++) {
						//Fetch the array index from precompute prefix sum of offset and current ocupation of given cell.
						int index		   = z*dimX*dimY + y*dimX + x;
						int cellOffset	   = grid[index].offset + offsets[index]++;	

						//Store the triangle reference
						sse::store(triangle,&cellTriangles[cellOffset]);
					};

			curFace++;
		};

		//Add vertex cound to global triangle index start
		vertexSum	 += object->vertexCount;
		vertexSumSse  = sse::set(0,vertexSum,vertexSum,vertexSum);
	};

	delete[] offsets;
	delete[] extents;

	#ifdef MEASURE_TIME
		#ifdef _WIN32
			QueryPerformanceCounter((LARGE_INTEGER*)&stop);
			printf("%f",(double)(stop-start)/TICKS_PER_SEC);
		#else
			stop = clock();
			printf("%f\n",(double)(stop-start)/CLOCKS_PER_SEC);
		#endif
	#endif
}

bool UniformGrid::prepareCudaTexture(float3 &minimum, float3 &maximum, float &cellSize, uint3 &dim,
									 cudaArray* &cudaGrid,uint4* &triangles,cudaArray* &vertices,cudaArray* &normalsOut) {
	dim.x = dimX;
	dim.y = dimY;
	dim.z = dimZ;

	cellSize = cuCellSize;
	minimum  = cuMinimum;
	maximum	 = cuMaximum;

	cudaGrid	= 0;
	vertices	= 0;
	normalsOut	= 0;
	triangles	= 0;

	//Allocate memory and copy data to video card. The actual exact size cannot be known in advance
	//because the device dependent padding. So just allocate the memory and check for failre. In that
	//case free the alreade allocated memory.

	try {
/////////////////////////////
// Copy grid cells to video memory as cudaArray, ready to be used for a texture.
// The array is 3D with the same size as the grid

	cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<uint2>();
	cudaExtent			  gridSize	  = make_cudaExtent(dimX, dimY, dimZ);

	SAFE_ALLOC(cudaMalloc3DArray(&cudaGrid, &channelDesc, gridSize));

	cudaMemcpy3DParms copyParams = {0};
	copyParams.srcPtr   = make_cudaPitchedPtr((void*)grid, gridSize.width*sizeof(uint2), gridSize.width, gridSize.height);
	copyParams.dstArray = cudaGrid;
    copyParams.extent   = gridSize;
    copyParams.kind     = cudaMemcpyHostToDevice;

	cudaMemcpy3D(&copyParams);

/////////////////////////////
//  Copy vertices and normals to the video memory. This is placed into texture as optimization.
//  The texture access is cached, unlike the main memory. The vertices have an unpredictable
//  access patern and cannot be coalesced.

	//The texture must be at least 2D, because the 1D size limit.
	unsigned sideX = 256;						
	unsigned sideY = totalVertex/sideX + 1;

	channelDesc = cudaCreateChannelDesc<float4>();
	gridSize	= make_cudaExtent(sideX, sideY, 0);

	SAFE_ALLOC(cudaMallocArray(&vertices, &channelDesc, sideX,sideY));
	cudaMemcpyToArray(vertices,0,0,transformed,totalVertex*sizeof(float4),cudaMemcpyHostToDevice);

	SAFE_ALLOC(cudaMallocArray(&normalsOut, &channelDesc, sideX,sideY));
	cudaMemcpyToArray(normalsOut,0,0,normals,totalVertex*sizeof(float4),cudaMemcpyHostToDevice);

/////////////////////////////

	SAFE_ALLOC(cudaMalloc((void**)&triangles,totalPoly*sizeof(Triangle)));
	cudaMemcpy(triangles,cellTriangles	,totalPoly*sizeof(Triangle)	,cudaMemcpyHostToDevice);

/////////////////////////////
	} catch(OutOfMemory) {
		if(triangles)  cudaFree(triangles);
		if(cudaGrid)   cudaFreeArray(cudaGrid);
		if(vertices)   cudaFreeArray(vertices);
		if(normalsOut) cudaFreeArray(normalsOut);
		throw OutOfMemory();
	};

	//Save the allocated pointer for freeing during object destruction.
	videoAllocated  = true;
	savedGrid		= cudaGrid;
	savedTriangles  = triangles;
	savedVertices	= vertices;
	savedNormals	= normalsOut;

	return true;
};

UniformGrid::~UniformGrid(void) {
	//Delete the host side of acceleration structure
	delete[] grid;
	delete[] transformed;
	delete[] cellTriangles;
	delete[] normals;

	if(videoAllocated) {	//Free video card memory if it was succesfully allocated.
		cudaFree(savedTriangles);
		cudaFreeArray(savedGrid);
		cudaFreeArray(savedVertices);
		cudaFreeArray(savedNormals);
	};
}
