/*
**	  GPU Raytrace - Diploma thesis, Czech technical university
**	  Copyright (C) 2008 Martin Zlatuska
**
**	  This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include <stdio.h>
#include <math.h>
#include <float.h>

#include <windows.h>
#include <vector_functions.h>

#include <cuda.h>

#include "Camera.h"
#include "SceneLights.h"
#include "Material.h"
#include "Partitoning/UniformGrid.h"
#include "Partitoning/KdTree.h"
#include "Partitoning/Bvh.h"

#include "Transformations/AnimationPath.h"
#include "Gpu/Wrapper.h"

#include "TestFunctions.h"

using std::max;
using std::min;

const float PI2  = 6.283185307179586476925286766559f;
const float PI2A = 57.295779513082320876798154814105f;
const float STEP = 0.17453292519943295769236907684886f;

FILE* logOut;

struct BitmapHeader {
	char		type[2];
	unsigned	size;
	short		reserve1;
	short		reserve2;
	unsigned	dataStart;
};

struct BitmapInfo {
	unsigned		size;
	unsigned		width;
	unsigned		height;
	unsigned short	planes;
	unsigned short	colors;
	unsigned		compression;
	unsigned		imageSize;
	unsigned		xppm;
	unsigned		yppm;
	unsigned		usedColors;
	unsigned		importantColors;
};

void writeBmp(unsigned* buffer,unsigned width,unsigned height,char* fileName) {
	BitmapHeader header;
	BitmapInfo   info;

	header.type[0] = 'B';
	header.type[1] = 'M';
	header.size	   =   14+sizeof(BitmapInfo)+width*height*3;
	header.reserve1 = 0;
	header.reserve2 = 0;
	header.dataStart = 14+sizeof(BitmapInfo);

	info.size			= 40;
	info.width			= width;
	info.height			= height;
	info.planes			= 1;
	info.colors			= 24;
	info.compression	= 0;
	info.imageSize		= width*height*3;
	info.xppm			= 0;
	info.yppm			= 0;
	info.usedColors		= 0;
	info.importantColors = 0;

	char* repack = new char[width*height*3];
	unsigned *cur = buffer + width*height-width;

	unsigned offset = 0;
	for(unsigned i=0;i<height;i++) {
		for(unsigned j=0;j<width;j++) {
			repack[offset++] = ((char*)cur)[2];
			repack[offset++] = ((char*)cur)[1];
			repack[offset++] = ((char*)cur)[0];
			cur++;
		};
		cur = buffer + width*(height-i)-width;
	};

	FILE *outout = fopen(fileName,"wb");

	fwrite(&header.type		,1,2,outout);
	fwrite(&header.size		,1,4,outout);
	fwrite(&header.reserve1	,1,2,outout);
	fwrite(&header.reserve2	,1,2,outout);
	fwrite(&header.dataStart,1,4,outout);
	fwrite(&info			,sizeof(info),1,outout);
	fwrite(repack,3,width*height,outout);

	fclose(outout);

	delete repack;
};

static float3 timingDirect[4][3];
static float3 timingLight [4][3];

void circleAround(const char* name,CudaRenderer *renderer,unsigned width,unsigned height,unsigned res,unsigned structure,float distRatio,bool save=false) {
	Camera camera;
	camera.setUpProjection(width,height,1.0f);

	renderer->resize(width,height);

	SceneLights lights;
	lights.push_back(Light());
	lights.back().possition = make_float3(0.0f,0.0f,10.0f);
	lights.back().kAmbient  = make_float3(0.2f,0.2f,0.2f);
	lights.back().kDiffuse  = make_float3(0.8f,0.8f,0.8f);

	float maxTime = 0.0f;
	float minTime = FLT_MAX;
	float sumTime = 0.0f;

	float angleX = 0.0f;
	float angleY = 0.0f;

	float distance = renderer->getRenderDistance()*distRatio;

#ifdef _WIN32
	HANDLE stdOut = GetStdHandle(STD_OUTPUT_HANDLE);
	CONSOLE_SCREEN_BUFFER_INFO screenInfo;
	GetConsoleScreenBufferInfo(stdOut,&screenInfo);
#endif

	unsigned frames = 0;

	for(int i=0;i<408;i++) {
		frames++;

		SPVector axis  (cosf(angleX),0.0f,-sinf(angleX),0.0f);
		SPVector eye   (distance * sinf(angleX),0.0f,distance * cosf(angleX),1.0f);
		SPVector up	   (0.0f,1.0f,0.0f,0.0f);
		SPVector target(0.0f,0.0f,0.0f,1.0f);

		Transformation rotate;
		rotate.IdentityMatrix();
		rotate.rotate(axis.x,axis.y,axis.z,angleY);

		eye *= rotate;
		up  *= rotate;
				
		camera.setLook(eye,target,up);
		lights.back().possition = make_float3(eye.x,eye.y,eye.z);

		char fileName[256];
		sprintf(fileName,"render\\%s-around-%03i.bmp",name,i);
		float time = renderer->render(camera,lights,PRIMARY);

#ifdef _WIN32
		SetConsoleCursorPosition(stdOut,screenInfo.dwCursorPosition);
		printf("%i%%",i*100/408);
#else
		printf("%i%%\n",i*100/408);
#endif

		sumTime+= time;
		maxTime = max(maxTime,time);
		minTime = min(minTime,time);

		angleX += 0.1f;
		angleY += 0.015f;

		if(angleX >= PI2) angleX -= PI2;
		if(angleY >= PI2) angleY -= PI2;
		
		if(save || i==0)
			writeBmp(renderer->getBuffer(),width,height,fileName);
	};

	timingDirect[res][structure].x = minTime;
	timingDirect[res][structure].y = maxTime;
	timingDirect[res][structure].z = sumTime/frames;

	printf("\n");
};

void circleLight(const char* name,CudaRenderer *renderer,unsigned width,unsigned height,unsigned res,unsigned structure,float distRatio,bool save=false) {
	Camera camera;
	camera.setUpProjection(width,height,1.0f);

	renderer->resize(width,height);

	SceneLights lights;
	lights.push_back(Light());
	lights.back().possition = make_float3(0.0f,0.0f,10.0f);
	lights.back().kAmbient  = make_float3(0.2f,0.2f,0.2f);
	lights.back().kDiffuse  = make_float3(0.8f,0.8f,0.8f);

	float maxTime = 0.0f;
	float minTime = FLT_MAX;
	float sumTime = 0.0f;

	float angleX = 0.0f;
	float angleY = 0.0f;

	float distance = renderer->getRenderDistance()*distRatio;

#ifdef _WIN32
	HANDLE stdOut = GetStdHandle(STD_OUTPUT_HANDLE);
	CONSOLE_SCREEN_BUFFER_INFO screenInfo;
	GetConsoleScreenBufferInfo(stdOut,&screenInfo);
#endif

	unsigned frames = 0;

	for(int i=0;i<408;i++) {
		frames++;

		SPVector eye   (0.0f,0.0f,distance,1.0f);
		SPVector target(0.0f,0.0f,0.0f,1.0f);
		SPVector up	   (0.0f,1.0f,0.0f,0.0f);

		camera.setLook(eye,target,up);
		
		SPVector axis  (cosf(angleX),0.0f,-sinf(angleX),0.0f);
		SPVector pos   (distance * sinf(angleX),0.0f,distance * cosf(angleX),1.0f);

		Transformation rotate;
		rotate.IdentityMatrix();
		rotate.rotate(axis.x,axis.y,axis.z,angleY);

		pos *= rotate;
				
		lights.back().possition = make_float3(pos.x,pos.y,pos.z);

		char fileName[256];
		sprintf(fileName,"render\\%s-light-%03i.bmp",name,i);
		float time = renderer->render(camera,lights,PRIMARY_SHADOW);

#ifdef _WIN32
		SetConsoleCursorPosition(stdOut,screenInfo.dwCursorPosition);
		printf("%i%%",i*100/408);
#else
		printf("%i%%\n",i*100/408);
#endif

		sumTime+= time;
		maxTime = max(maxTime,time);
		minTime = min(minTime,time);

		angleX += 0.1f;
		angleY += 0.015f;

		if(angleX >= PI2) angleX -= PI2;
		if(angleY >= PI2) angleY -= PI2;
		
		if(save)
			writeBmp(renderer->getBuffer(),width,height,fileName);
	};

	timingLight[res][structure].x = minTime;
	timingLight[res][structure].y = maxTime;
	timingLight[res][structure].z = sumTime/frames;

	printf("\n");
};

unsigned resX[]={256,512,768,1024};
unsigned resY[]={256,512,768,1024};
bool doRes[4];

void resTest(const char* name,CudaRenderer *renderer,float distRatio) {
	Camera camera;

	SceneLights lights;
	lights.push_back(Light());
	lights.back().possition = make_float3(0.0f,0.0f,10.0f);
	lights.back().kAmbient  = make_float3(0.2f,0.2f,0.2f);
	lights.back().kDiffuse  = make_float3(0.8f,0.8f,0.8f);

	float maxTime = 0.0f;
	float minTime = FLT_MAX;
	float sumTime = 0.0f;

	float angleX = 5.43f;
	float angleY = 2.7f;

	float distance = renderer->getRenderDistance()*distRatio;

	unsigned baseX=512;
	unsigned baseY=512;

	SPVector axis  (cosf(angleX),0.0f,-sinf(angleX),0.0f);
	SPVector eye   (distance * sinf(angleX),0.0f,distance * cosf(angleX),1.0f);
	SPVector up	   (0.0f,1.0f,0.0f,0.0f);
	SPVector target(0.0f,0.0f,0.0f,1.0f);

	Transformation rotate;
	rotate.IdentityMatrix();
	rotate.rotate(axis.x,axis.y,axis.z,angleY);

	eye *= rotate;
	up  *= rotate;
				
	lights.back().possition = make_float3(eye.x,eye.y,eye.z);
	
	char blabla[128];

	float fov=0.8f;

	for(int i=0;i<33;i++) {
		renderer->resize(baseX,baseY);

		camera.setUpProjection(baseX,baseY,1.0f);
		camera.setLook(eye,target,up);
	
		float time = renderer->render(camera,lights,PRIMARY);
		printf("%ix%i %f ms\n",baseX,baseY,time);

		sprintf(blabla,"render/test%02i.bmp",i);
		writeBmp(renderer->getBuffer(),baseX,baseY,blabla);

		baseX+=16;
		baseY+=16;
	};
};

void doSimpleTest(char *filename,char *name,float distance) {
	char tempName[256];

	Scene* scene = SceneLoader::instance()->loadScene(filename,true);
	CudaRenderer* renderer;

	UniformGrid* grid = new UniformGrid(*scene);
	renderer = new CudaRenderGrid(512,512,grid);

	printf("%s\n",name);

	for(unsigned i=0;i<4;i++) {
		if(doRes[i]) {
			renderer->resize(resX[i],resY[i]);

			printf("GRID PRIMARY %i %i ",resX[i],resY[i]);

			sprintf(tempName,"%s-grid",name);
			circleAround(tempName,renderer,resX[i],resY[i],i,0,distance,false);
		};
	};

	for(unsigned i=0;i<4;i++) {
		if(doRes[i]) {
			renderer->resize(resX[i],resY[i]);

			printf("GRID LIGHT %i %i ",resX[i],resY[i]);

			sprintf(tempName,"%s-grid",name);
			circleLight(tempName,renderer,resX[i],resY[i],i,0,distance,false);
		};
	};

	delete renderer;

	KdTree*			tree = new KdTree(*scene);
	renderer = new CudaRenderKd(512,512,tree);

	printf("%s\n",name);

	for(unsigned i=0;i<4;i++) {
		if(doRes[i]) {
			renderer->resize(resX[i],resY[i]);

			printf("KD-TREE PRIMARY %i %i ",resX[i],resY[i]);

			sprintf(tempName,"%s-kd",name);
			circleAround(tempName,renderer,resX[i],resY[i],i,1,distance,false);
		};
	};

	for(unsigned i=0;i<4;i++) {
		if(doRes[i]) {
			renderer->resize(resX[i],resY[i]);

			printf("KD-TREE LIGHT %i %i ",resX[i],resY[i]);

			sprintf(tempName,"%s-kd",name);
			circleLight(tempName,renderer,resX[i],resY[i],i,1,distance,false);
		};
	};

	delete renderer;

	Bvh* btree = new Bvh(*scene);
	renderer = new CudaRenderBvh(512,512,btree);

	printf("%s\n",name);

	for(unsigned i=0;i<4;i++) {
		if(doRes[i]) {
			renderer->resize(resX[i],resY[i]);

			printf("BVH-TREE PRIMARY %i %i ",resX[i],resY[i]);

			sprintf(tempName,"%s-bvh",name);
			circleAround(tempName,renderer,resX[i],resY[i],i,2,distance,false);
		};
	};

	for(unsigned i=0;i<4;i++) {
		if(doRes[i]) {
			renderer->resize(resX[i],resY[i]);

			printf("BVH-TREE LIGHT %i %i ",resX[i],resY[i]);

			sprintf(tempName,"%s-bvh",name);
			circleLight(tempName,renderer,resX[i],resY[i],i,2,distance,false);
		};
	};

	delete renderer;

	delete scene;

	fprintf(logOut,"%s\n",name);
	fprintf(logOut,",primary\n");
	fprintf(logOut,",GRID,,,,KD-TREE,,,,BVH\n");

	for(int i=0;i<4;i++)
		if(doRes[i]) {
			fprintf(logOut,"%ix%i,",resX[i],resY[i]);
			fprintf(logOut,",%f,%f,%f,,%f,%f,%f,,%f,%f,%f\n",	timingDirect[i][0].x,timingDirect[i][0].y,timingDirect[i][0].z,
																timingDirect[i][1].x,timingDirect[i][1].y,timingDirect[i][1].z,
																timingDirect[i][2].x,timingDirect[i][2].y,timingDirect[i][2].z);
		}

	fprintf(logOut,",light\n");
	fprintf(logOut,",GRID,,,,KD-TREE,,,,BVH\n");

	for(int i=0;i<4;i++)
		if(doRes[i]) {
			fprintf(logOut,"%ix%i,",resX[i],resY[i]);
			fprintf(logOut,",%f,%f,%f,,%f,%f,%f,,%f,%f,%f\n",	timingLight[i][0].x,timingLight[i][0].y,timingLight[i][0].z,
																timingLight[i][1].x,timingLight[i][1].y,timingLight[i][1].z,
																timingLight[i][2].x,timingLight[i][2].y,timingLight[i][2].z);
		};

	fflush(logOut);
};

void prepareTest(const char* filename) {
	logOut = fopen(filename,"w");
	if(!logOut) {
		printf("Program must be run from writable directory\n");
		exit(1);
	};

	printf("Select resolutions:\n0- All\n1 - 256x256\n2 - 512x512\n3 - 768x768\n4 - 1024x1024\n");
	printf("Multiple choices accepted\n");

	doRes[0] = doRes[1] = doRes[2] = doRes[3] = false;

	int  res[4];
	char buffer[256];

	fgets(buffer,256,stdin);
	fgets(buffer,256,stdin);
	int count = sscanf(buffer,"%i %i %i %i %i",&res[0],&res[1],&res[2],&res[3],&res[4]);
	//int count=0;
	//res[0]=0;

	if(res[0] == 0)
		doRes[0] = doRes[1] = doRes[2] = doRes[3] = true;
	else {
		for(int i=0;i<count;i++) {
			int curRes = res[i];
			if(curRes >= 1 && curRes <=5)
				doRes[curRes-1]=true;
		};
	};
};

void endTest() {
	fclose(logOut);
};


void structStats(const char* filename,const char* name,unsigned cellsize) {
	UniformGrid*	grid;
	KdTree*			kdTree;
	Bvh*			bvhTree;

	Scene*			scene;

	scene = SceneLoader::instance()->loadScene(filename);
	/*grid	= new UniformGrid(*scene,cellsize);

	unsigned totalReference = 0;
	unsigned maxCount	    = 0;
	
	for(unsigned i=0;i<grid->dimX*grid->dimY*grid->dimZ;i++) {
		totalReference += grid->grid[i].count;
		maxCount = max(maxCount,grid->grid[i].count);
	};

	unsigned structSize = grid->dimX*grid->dimY*grid->dimZ*8 +	grid->totalPoly*sizeof(Triangle);
	unsigned modelSize  = grid->totalVertex*2*16;
	unsigned total		= structSize + modelSize;

	printf("%s,%ix%ix%i,%i,%.2f,%.2f,%.2f\n",name,grid->dimX,grid->dimY,grid->dimZ,maxCount,(float)totalReference/(*scene)[0].object->facesCount,(float)structSize/1024/1024,(float)total/1024/1024);
	delete grid;*/

	kdTree	= new KdTree(*scene);

	unsigned totalReference = 0;
	unsigned maxCount	    = 0;
	unsigned leafs			= 0;
	unsigned emptyLeafs		= 0;

	unsigned maxIndex=0;

	unsigned histo[256];
	memset(histo,0,256*4);

	for(unsigned i=0;i<kdTree->nodes.size();i++) {
		if((kdTree->nodes[i]->flags & 0x3) == LEAF) {
			totalReference += kdTree->nodes[i]->count>>2;

			if((kdTree->nodes[i]->count>>2) > maxCount)
				maxIndex=i;

			histo[kdTree->nodes[i]->count>>2]++;
			
			maxCount = max(maxCount,kdTree->nodes[i]->count>>2);

			leafs++;
			if(kdTree->nodes[i]->count>>2 == 0) emptyLeafs++;
		};
	};

	unsigned structSize = kdTree->nodes.size()*8 +	kdTree->nodeTriangles.size()*sizeof(Triangle);
	unsigned modelSize  = kdTree->vertexSum*2*16;
	unsigned total		= structSize + modelSize;

	printf("%s,%i,%i,%i,%.2f,%.2f,%.2f\n",name,leafs,emptyLeafs,maxCount,(float)totalReference/(*scene)[0].object->facesCount,(float)structSize/1024/1024,(float)total/1024/1024);
	delete kdTree;
/*
	bvhTree = new Bvh(*scene);

	unsigned structSize = bvhTree->cells->size()*8 +	bvhTree->cellTriangles->size()*sizeof(Triangle);
	unsigned modelSize  = bvhTree->vertexSum*2*16;
	unsigned total		= structSize + modelSize;

	printf("%s,%i,%.2f\n",name,bvhTree->cells->size(),(float)total/1024/1024);
	delete bvhTree;
	*/
	delete scene;
};