/*
**	  GPU Raytrace - Diploma thesis, Czech technical university
**	  Copyright (C) 2008 Martin Zlatuska
**
**	  This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#pragma once

void writeBmp(unsigned* buffer,unsigned width,unsigned height,char* fileName);

void doSimpleTest(char *filename,char *name,float distance);
void doGridTest(char *filename,char *name,float distance);

void prepareTest(const char* filename);
void endTest();

void structStats(const char* filename,const char* name,unsigned cellsize=-1);

#include <direct.h>

template<typename LightOps>
void walkScene(char* filename, char* pathname,unsigned resolution,int gridSize,bool recenter,LightOps lightOps) {
	extern unsigned resX[];
	extern unsigned resY[];
	extern bool doRes[];

	extern FILE* logOut;

	static const char* STRUCT_NAME[] = {"Grid","KD","BVH"};
	static const char* RENDER_NAME[] = {"Primary","Shadow","Reflect"};

	float3 timing[3][4][3];

	char sceneName[256];
	char dirName[256];

#ifdef _WIN32
	HANDLE stdOut = GetStdHandle(STD_OUTPUT_HANDLE);
	CONSOLE_SCREEN_BUFFER_INFO screenInfo;
#endif

	unsigned length;
	char* start = strrchr(filename,'/');
	if(start==0) start = filename; else start++;
	char* end   = strrchr(filename,'.');
	if(end) length = end-start; else length = strlen(start);
	strncpy(sceneName,start,length);
	sceneName[length]=0;

	AnimationPath path(pathname);

	Scene* scene = SceneLoader::instance()->loadScene(filename,recenter);
		
	SceneLights lights;
	lightOps.setup(lights);

	float maxTime,minTime,sumTime;

	printf("%s\n",sceneName);

	sprintf(dirName,"render\\%s",sceneName);
	_mkdir(dirName);

	for(unsigned j=0;j<3;j++) {
		CudaRenderer *renderer;
		switch(j) {
			case 0: { UniformGrid* grid = new UniformGrid(*scene,gridSize);
				      renderer			= new CudaRenderGrid(resX[resolution],resY[resolution],grid); break; }
			case 1: { KdTree* tree		= new KdTree(*scene);
					  renderer			= new CudaRenderKd(resX[resolution],resY[resolution],tree);   break; }
			case 2:	{ Bvh* tree			= new Bvh(*scene);
					  renderer			= new CudaRenderBvh(resX[resolution],resY[resolution],tree);  break; }
		};

		for(unsigned res=0;res<4;res++) {
			if(!doRes[res]) continue;
			
			printf("%ix%i",resX[res],resY[res]);

			Camera camera;
			camera.setUpProjection(resX[res],resY[res],1.0f);

			renderer->resize(resX[res],resY[res]);

			for(unsigned k=0;k<3;k++) {
				if(resolution && j==1 && k==2) continue;

				printf("\t%s-%s:",STRUCT_NAME[j],RENDER_NAME[k]);

				GetConsoleScreenBufferInfo(stdOut,&screenInfo);

				bool gotPos,gotDir;
				float3 pos;
				float3 viewDir,viewUp;

				float timePos = path.start;
				float step    = (path.end - path.start)/path.framesCount;

				maxTime = 0.0f;
				minTime = FLT_MAX;
				sumTime = 0.0f;

				unsigned realFrames = 0;
				for(int i=0;i<path.framesCount;i++) {
					path.GetCamera(timePos,gotPos,pos,gotDir,viewDir,viewUp);
					timePos += step;

					if(i%20) continue;

					realFrames++;

					SPVector eye	(pos.x,pos.y,pos.z,1.0f);
					SPVector target (pos.x + viewDir.x,pos.y + viewDir.y,pos.z + viewDir.z);
					SPVector up		(viewUp.x,viewUp.y,viewUp.z);
					camera.setLook(eye,target,up);

					lightOps(lights,pos);

					char outName[256];
					sprintf(outName,"render\\%s\\%s-%s-%03i.bmp",sceneName,STRUCT_NAME[j],RENDER_NAME[k],i);
					float time = renderer->render(camera,lights,(RenderMode)k);

					SetConsoleCursorPosition(stdOut,screenInfo.dwCursorPosition);
					printf("%i%%",i*100/path.framesCount);
					//printf("%i:%f ms\n",i,time);

					sumTime+= time;
					maxTime = max(maxTime,time);
					minTime = min(minTime,time);
					
					//writeBmp(renderer->getBuffer(),resX[res],resY[res],outName);
				};

				timing[j][res][k].x = minTime;
				timing[j][res][k].y = maxTime;
				timing[j][res][k].z = sumTime/realFrames;
				printf("\n");
			};
		};
		delete renderer;
	};
	delete scene;

	fprintf(logOut,"%s %ix%i\n",sceneName,resX[resolution],resY[resolution]);
	fprintf(logOut,",primary\n");
	fprintf(logOut,",GRID,,,,KD-TREE,,,,BVH\n");

	for(unsigned k=0;k<4;k++) {
		if(doRes[k]) {
			fprintf(logOut,"%ix%i,",resX[k],resY[k]);
			fprintf(logOut,"%f,%f,%f,,%f,%f,%f,,%f,%f,%f\n",	timing[0][k][0].x,timing[0][k][0].y,timing[0][k][0].z,
																timing[1][k][0].x,timing[1][k][0].y,timing[1][k][0].z,
																timing[2][k][0].x,timing[2][k][0].y,timing[2][k][0].z);
		};
	};

	fprintf(logOut,",shadow\n");
	fprintf(logOut,",GRID,,,,KD-TREE,,,,BVH\n");

	for(unsigned k=0;k<4;k++) {
		if(doRes[k]) {
			fprintf(logOut,"%ix%i,",resX[k],resY[k]);
			fprintf(logOut,"%f,%f,%f,,%f,%f,%f,,%f,%f,%f\n",	timing[0][k][1].x,timing[0][k][1].y,timing[0][k][1].z,
																timing[1][k][1].x,timing[1][k][1].y,timing[1][k][1].z,
																timing[2][k][1].x,timing[2][k][1].y,timing[2][k][1].z);
		};
	};

	fprintf(logOut,",reflect\n");
	fprintf(logOut,",GRID,,,,KD-TREE,,,,BVH\n");

	for(unsigned k=0;k<4;k++) {
		if(doRes[k]) {
			fprintf(logOut,"%ix%i,",resX[k],resY[k]);
			fprintf(logOut,"%f,%f,%f,,%f,%f,%f,,%f,%f,%f\n",	timing[0][k][2].x,timing[0][k][2].y,timing[0][k][2].z,
																timing[1][k][2].x,timing[1][k][2].y,timing[1][k][2].z,
																timing[2][k][2].x,timing[2][k][2].y,timing[2][k][2].z);
		};
	};
	fflush(logOut);
};

#if 0
template<typename LightOps>
void walkScene(char* filename, char* pathname,unsigned resolution,int gridSize,bool recenter,LightOps lightOps) {
	extern unsigned resX[];
	extern unsigned resY[];

	extern FILE* logOut;

	static const char* STRUCT_NAME[] = {"Grid","KD","BVH"};
	static const char* RENDER_NAME[] = {"Primary","Shadow","Reflect"};

	float3 timing[3][4][3];

	char sceneName[256];
	char dirName[256];

	HANDLE stdOut = GetStdHandle(STD_OUTPUT_HANDLE);
	CONSOLE_SCREEN_BUFFER_INFO screenInfo;

	unsigned length;
	char* start = strrchr(filename,'/');
	if(start==0) start = filename; else start++;
	char* end   = strrchr(filename,'.');
	if(end) length = end-start; else length = strlen(start);
	strncpy(sceneName,start,length);
	sceneName[length]=0;

	AnimationPath path(pathname);

	Scene* scene = SceneLoader::instance()->loadScene(filename,recenter);
		
	SceneLights lights;
	lightOps.setup(lights);

	float maxTime,minTime,sumTime;

	printf("%s\n",sceneName);

	sprintf(dirName,"render\\%s",sceneName);
	_mkdir(dirName);

	//for(unsigned j=0;j<3;j++) {
	for(unsigned j=1;j<2;j++) {
		CudaRenderer *renderer;
		switch(j) {
			case 0: { UniformGrid* grid = new UniformGrid(*scene,gridSize);
				      renderer			= new CudaRenderGrid(resX[resolution],resY[resolution],grid); break; }
			case 1: { KdTree* tree		= new KdTree(*scene);
					  renderer			= new CudaRenderKd(resX[resolution],resY[resolution],tree);   break; }
			case 2:	{ Bvh* tree			= new Bvh(*scene);
					  renderer			= new CudaRenderBvh(resX[resolution],resY[resolution],tree);  break; }
		};

		unsigned baseX=512;
		unsigned baseY=512;

		//for(unsigned res=0;res<4;res++) {
		for(unsigned res=0;res<40;res++) {
		//for(unsigned res=0;res<128;res++) {
			/*printf("%ix%i",resX[res],resY[res]);

			Camera camera;
			camera.setUpProjection(resX[res],resY[res],1.0f);

			renderer->resize(resX[res],resY[res]);*/

			printf("%ix%i\t",baseX,baseY);
			//printf("%f\t",sqrtf(length*64));
			//printf("%i\t",baseX);

			Camera camera;
			camera.setUpProjection(baseX,baseY,1.0f);
			//camera.setUpProjection(1024,1024,1.0f);

			renderer->resize(baseX,baseY);

			//for(unsigned k=0;k<3;k++) {
			for(unsigned k=0;k<1;k++) {
				//printf("\t%s-%s:",STRUCT_NAME[j],RENDER_NAME[k]);

				GetConsoleScreenBufferInfo(stdOut,&screenInfo);

				bool gotPos,gotDir;
				float3 pos;
				float3 viewDir,viewUp;

				float timePos = path.start;
				float step    = (path.end - path.start)/path.framesCount;

				maxTime = 0.0f;
				minTime = FLT_MAX;
				sumTime = 0.0f;

				unsigned realFrames = 0;
				for(int i=0;i<path.framesCount;i++) {
				//for(int i=0;i<1;i++) {
					path.GetCamera(timePos,gotPos,pos,gotDir,viewDir,viewUp);
					timePos += step;

					//if(i!=500) continue;
					//if(i%10!=0) continue;
					//if(i%2!=0) continue;
					if(i!=130) continue;

					realFrames++;

					SPVector eye	(pos.x,pos.y,pos.z,1.0f);
					SPVector target (pos.x + viewDir.x,pos.y + viewDir.y,pos.z + viewDir.z);
					SPVector up		(viewUp.x,viewUp.y,viewUp.z);
					camera.setLook(eye,target,up);

					lightOps(lights,pos);

					char outName[256];
					//sprintf(outName,"render\\%s\\%s-%s-%03i.bmp",sceneName,STRUCT_NAME[j],RENDER_NAME[k],i);
					sprintf(outName,"render\\test%02i.bmp",res);
					//sprintf(outName,"render\\test%02i.raw",res);
					float time = renderer->render(camera,lights,(RenderMode)k);

					//SetConsoleCursorPosition(stdOut,screenInfo.dwCursorPosition);
					//printf("%i%%",i*100/path.framesCount);
					//printf("%i:%f ms\n",i,time);

					sumTime+= time;
					maxTime = max(maxTime,time);
					minTime = min(minTime,time);
					
					//if((res==2 && j==2) || i==0)
					//	writeBmp(renderer->getBuffer(),resX[res],resY[res],outName);
					writeBmp(renderer->getBuffer(),baseX,baseY,outName);
					/*unsigned sum=0;
					unsigned* data=renderer->getBuffer();
					for(unsigned bla=0;bla<baseX*baseY;bla++)
						sum+=data[bla];*/

					//printf(" %i ",sum);

					/*FILE* output;
					output=fopen(outName,"wb");
					fwrite(renderer->getBuffer(),4,baseX*baseY,output);
					fclose(output);*/
				};

				/*timing[j][res][k].x = minTime;
				timing[j][res][k].y = maxTime;
				timing[j][res][k].z = sumTime/realFrames;
				//printf("minTime: %f\nmaxTime: %f\naverage: %f\n",minTime,maxTime,sumTime/path.framesCount);
				printf("\n");*/
				printf(" %f\n",maxTime);
			};

			if(baseX<768)
				baseX+=16;
			else
				baseX+=32;

			if(baseY<768)
				baseY+=16;
			else
				baseY+=32;
		};

		delete renderer;
	};

	delete scene;

	/*fprintf(logOut,"%s %ix%i\n",sceneName,resX[resolution],resY[resolution]);
	fprintf(logOut,",primary\n");
	fprintf(logOut,",GRID,,,,KD-TREE,,,,BVH\n");

	for(unsigned k=0;k<4;k++) {
		fprintf(logOut,",%f,%f,%f,,%f,%f,%f,,%f,%f,%f\n",	timing[0][k][0].x,timing[0][k][0].y,timing[0][k][0].z,
															timing[1][k][0].x,timing[1][k][0].y,timing[1][k][0].z,
															timing[2][k][0].x,timing[2][k][0].y,timing[2][k][0].z);
	};

	fprintf(logOut,",shadow\n");
	fprintf(logOut,",GRID,,,,KD-TREE,,,,BVH\n");

	for(unsigned k=0;k<3;k++) {
		fprintf(logOut,",%f,%f,%f,,%f,%f,%f,,%f,%f,%f\n",	timing[0][k][1].x,timing[0][k][1].y,timing[0][k][1].z,
															timing[1][k][1].x,timing[1][k][1].y,timing[1][k][1].z,
															timing[2][k][1].x,timing[2][k][1].y,timing[2][k][1].z);
	};

	fprintf(logOut,",reflect\n");
	fprintf(logOut,",GRID,,,,KD-TREE,,,,BVH\n");

	for(unsigned k=0;k<3;k++) {
		fprintf(logOut,",%f,%f,%f,,%f,%f,%f,,%f,%f,%f\n",	timing[0][k][2].x,timing[0][k][2].y,timing[0][k][2].z,
															timing[1][k][2].x,timing[1][k][2].y,timing[1][k][2].z,
															timing[2][k][2].x,timing[2][k][2].y,timing[2][k][2].z);
	};

	

	fflush(logOut);*/
};
#endif

template<typename LightOps>
void walkGrid(char* filename, char* pathname,bool recenter,LightOps lightOps) {
	extern FILE* logOut;

	float3 timing[3];

	char sceneName[256];

	HANDLE stdOut = GetStdHandle(STD_OUTPUT_HANDLE);
	CONSOLE_SCREEN_BUFFER_INFO screenInfo;

	unsigned length;
	char* start = strrchr(filename,'/');
	if(start==0) start = filename; else start++;
	char* end   = strrchr(filename,'.');
	if(end) length = end-start; else length = strlen(start);
	strncpy(sceneName,start,length);
	sceneName[length]=0;

	AnimationPath path(pathname);

	Scene* scene = SceneLoader::instance()->loadScene(filename,recenter);
		
	Camera camera;
	camera.setUpProjection(512,512,1.0f);

	SceneLights lights;
	lightOps.setup(lights);

	float maxTime,minTime,sumTime;

	printf("%s\n",sceneName);
	fprintf(logOut,"s\n",sceneName);

	static const unsigned cells[]={24,32,48,64,86,92,128};
	static const char* RENDER_NAME[] = {"Primary","Shadow","Reflect"};

	for(unsigned j=1;j<7;j++) {
		CudaRenderer *renderer;
		
		UniformGrid* grid = new UniformGrid(*scene,cells[j]);
		renderer		  = new CudaRenderGrid(512,512,grid);
		
		for(unsigned k=0;k<1;k++) {
			printf("\t%s:",RENDER_NAME[k]);

			GetConsoleScreenBufferInfo(stdOut,&screenInfo);

			bool gotPos,gotDir;
			float3 pos;
			float3 viewDir,viewUp;

			float timePos = path.start;
			float step    = (path.end - path.start)/path.framesCount;

			maxTime = 0.0f;
			minTime = FLT_MAX;
			sumTime = 0.0f;

			unsigned realFrames =0;

			for(int i=0;i<path.framesCount;i++) {
			//for(int i=0;i<1;i++) {
				path.GetCamera(timePos,gotPos,pos,gotDir,viewDir,viewUp);
				timePos += step;

				//if(i!=500) continue;
				if(i%10!=0) continue;
				//if(i%2!=0) continue;
				//if(i!=150) continue;

				realFrames++;

				SPVector eye	(pos.x,pos.y,pos.z,1.0f);
				SPVector target (pos.x + viewDir.x,pos.y + viewDir.y,pos.z + viewDir.z);
				SPVector up		(viewUp.x,viewUp.y,viewUp.z);
				camera.setLook(eye,target,up);

				lightOps(lights,pos);

				char outName[256];
				sprintf(outName,"render\\%s-Grid-%s-%03i.bmp",sceneName,RENDER_NAME[k],i);
				float time = renderer->render(camera,lights,(RenderMode)k);

				SetConsoleCursorPosition(stdOut,screenInfo.dwCursorPosition);
				printf("%i%%",i*100/path.framesCount);
				//printf("%i:%f ms\n",i,time);

				sumTime+= time;
				maxTime = max(maxTime,time);
				minTime = min(minTime,time);
				
				writeBmp(renderer->getBuffer(),512,512,outName);
			};

			timing[k].x = minTime;
			timing[k].y = maxTime;
			timing[k].z = sumTime/realFrames;
			//printf("minTime: %f\nmaxTime: %f\naverage: %f\n",minTime,maxTime,sumTime/path.framesCount);
			printf("\n");
		};

		delete renderer;

		fprintf(logOut,"Cells:%i\n",cells[j]);
		fprintf(logOut,"Primary,%f,%f,%f\n", timing[0].x,timing[0].y,timing[0].z);
		fprintf(logOut,"Shadow ,%f,%f,%f\n", timing[1].x,timing[1].y,timing[1].z);
		fprintf(logOut,"Reflect,%f,%f,%f\n", timing[2].x,timing[2].y,timing[2].z);
		fflush(logOut);
	};
	delete scene;
};
