/*
**	  GPU Raytrace - Diploma thesis, Czech technical university
**	  Copyright (C) 2008 Martin Zlatuska
**
**	  This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#pragma once

#include <vector>

using std::vector;

class OutOfMemory {};
#define SAFE_ALLOC(result) if(result != cudaSuccess) throw OutOfMemory();

//! Simple growing array
/**
*	Somewhat similar to std::vector but using realloc function on out of capacity, so growing might not require
*	data copy. It doe's not support iterators but data placed in this array will usualy be copied to 
*	video card memory, where iterator cannot be used anyway and using index for all operations is therefore 
*	more benefitial.
*/
template<typename T>
class GrowArray {
	unsigned count;			//!< Number of currently used elements.
	unsigned allocated;		//!< Number of currently alocated elements.

	T*  data;				//!< Array of elements.

	//unsigned allocs;	//Used as a debug info for properly scaling the initial size estimate.

public:
	GrowArray(unsigned start = 512); 
	~GrowArray() {free(data);};

	//! Request one new elements. Return as index to array.
	unsigned getNext();

	//! Request several new elements. Return index to array of the first element.
	unsigned getMore(unsigned number);

	//! Return elemnt at given index. The return pointer can be invalid after subsequent memory requests.
	T* operator[](unsigned index) {return &data[index];};

	//! Return number of used elements.
	size_t size() {return count;};
};

template<typename T>
GrowArray<T>::GrowArray(unsigned start) {
	//allocs = 0;
	allocated = start;
	data = (T*) malloc(start * sizeof(T));
	count = 0;
};

template<typename T>
unsigned GrowArray<T>::getNext() {
	if(count == allocated) {
//		allocs++;
		allocated*=2;
		data = (T*)realloc(data,allocated * sizeof(T));
	};
	return count++;
};

template<typename T>
unsigned GrowArray<T>::getMore(unsigned number) {
	if(count + number > allocated) {
//		allocs++;
		allocated*=2;
		data = (T*)realloc(data,allocated * sizeof(T));
	};
	unsigned oldCount = count;
	count += number;
	return oldCount;
};


//! Simple stack-like pool.
/**
*	A simple pool that requires the memory to be returned in same order as it was requested.
*	No checking of this condition is done by the pool. \n
*	The pool allocates memory in chunks. New chunk is allocated if the currently used cannot
*	satisfy the request. The unused memory in current chunk is than waster. The maximum possible
*   request is passed as constructor parameter. \n
*	When memory is returned, no memory is dealocated. Insted it is marked as free and ready for
*	next requests. The allocated memory is freed only during the object destruction.
*/
template<typename T>
class TailPool {
	struct Chunk {
		T*		 data;
		unsigned used;

		Chunk(T* data):data(data),used(0) {};
	};

	vector<Chunk> chunks;	//!< The array of all allocated chunks.
	unsigned step;			//!< The size(in elements) of each allocation

	Chunk*	 chunk;			//!< Pointer to the current chunk
	unsigned chunkIndex;	//!< The index of current chunk in the chunk array
	
public:
	TailPool(unsigned step);
	~TailPool();

	//! Get new memory
	/**
	*	Returns pointer to linear array of elements with requested size. The requested count must not be larger than
	*   the value passed to constructor.
	*	@param count Size of requested memory in element count
	*/
	T*   getMem(unsigned count=1);	

	//! Return allocated memory.
	/** 
	*	Return the previously requested memory. Must be in returned in same order as requested!!!. No memory is actualy deallocated.
	*	@param count Size of returned memory in element count.
	*/
	void returnMem(unsigned count);

	void reset();
};

template<typename T>
TailPool<T>::TailPool(unsigned step):step(step) {
	chunkIndex = 0;

	//Automaticaly allocate first chunk
	chunks.push_back(Chunk(new T[step]));
	chunk = &chunks.front();
};

template<typename T>
TailPool<T>::~TailPool() {
	//Free all the allocated chunks memory.
	for(unsigned i=0;i<chunks.size();i++)
		delete[] chunks[i].data;

};

template<typename T>
void TailPool<T>::reset() {
	//Must mark all alocated chunk as free and reset current chunk pointer to first allocated.

	for(unsigned i=0;i<chunks.size();i++) 
		chunks[i].used =0;

	chunkIndex = 0;
	chunk	   = &chunks.front();
};

template<typename T>
T* TailPool<T>::getMem(unsigned count) {
	unsigned index = chunk->used;

	//Check if free space in chunk is enough to satisfy the request.
	if(chunk->used + count > step) {
		//It is not.
		chunkIndex++;

		//We might have unused allocated chunk ready if the user return some of the memory.
		if(chunks.size() > chunkIndex) { //We actualy have one, just update the current chunk pointer.
			chunk = &chunks[chunkIndex];
		} else { //No free chunk. Allocated new one.
			chunks.push_back(Chunk(new T[step]));
			chunk = &chunks.back();
		};

		index = 0;
		chunk->used = count;
	} else
		//It is. Just update used info in chunk
		chunk->used += count;

	//Return the first free address
	return &chunk->data[index];
};

template<typename T>
void TailPool<T>::returnMem(unsigned count) {
	chunk->used -= count;	
	if(chunk->used == 0 && chunkIndex > 0) {
		//If the current chunk is now compelete free, move to the previous one and keep this one for further use.
		chunkIndex--;
		chunk = &chunks[chunkIndex];
	};
};
